import os
from textwrap import dedent

from .cluster_manager import ClusterManager


class EDMClusterManager(ClusterManager):

    # The path to edm root on the remote, this is a relative path
    # and is relative to the home directory.
    EDM_ROOT = '.edm'

    ENV_FILE = "bundled_env.json"

    BOOTSTRAP = dedent("""\
    #!/bin/bash

    set -e
    ENV_FILE="{project_name}/{env_file}"

    if hash edm 2>/dev/null; then
        EDM_EXE=edm
    else
        EDM_EXE=~/{edm_root}/bin/edm
    fi

    if [ -f $ENV_FILE ] ; then
        $EDM_EXE -q envs import --force {project_name} -f $ENV_FILE
    else
        $EDM_EXE -q envs create --force {project_name} --version 3.6
        $EDM_EXE -q install psutil execnet -y -e {project_name}
    fi

    $EDM_EXE run -e {project_name} -- pip install automan

    cd {project_name}
    if [ -f "requirements.txt" ] ; then
        $EDM_EXE run -e {project_name} -- pip install -r requirements.txt
    fi
    """)

    UPDATE = dedent("""\
    #!/bin/bash

    set -e
    ENV_FILE="{project_name}/{env_file}"

    if hash edm 2>/dev/null; then
        EDM_EXE=edm
    else
        EDM_EXE=~/{edm_root}/bin/edm
    fi

    if [ -f $ENV_FILE ] ; then
        $EDM_EXE -q envs import --force {project_name} -f $ENV_FILE
    fi

    cd {project_name}
    if [ -f "requirements.txt" ] ; then
        $EDM_EXE run -e {project_name} -- pip install -r requirements.txt
    fi
    """)

    def _get_bootstrap_code(self):
        return self.BOOTSTRAP.format(
            project_name=self.project_name, edm_root=self.EDM_ROOT,
            env_file=self.ENV_FILE
        )

    def _get_python(self, host, home):
        return os.path.join(
            home, self.EDM_ROOT,
            'envs/{project_name}/bin/python'.format(
                project_name=self.project_name
            )
        )

    def _get_update_code(self):
        return self.UPDATE.format(
            project_name=self.project_name, edm_root=self.EDM_ROOT,
            env_file=self.ENV_FILE
        )

    def _get_helper_scripts(self):
        """Return a space separated string of script files that you need copied over to
        the remote host.

        When overriding this, you can return None or '' if you do not need any.

        """
        return ''



import os
from textwrap import dedent

from .cluster_manager import ClusterManager


class CondaClusterManager(ClusterManager):

    # The path to conda root on the remote, this is a relative path
    # and is relative to the home directory.
    CONDA_ROOT = 'miniconda3'

    BOOTSTRAP = dedent("""\
    #!/bin/bash

    set -e
    CONDA_ROOT={conda_root}
    ENV_FILE="{project_name}/environments.yml"
    if [ -f $ENV_FILE ] ; then
        ~/$CONDA_ROOT/bin/conda env create -q -f $ENV_FILE -n {project_name}
    else
        ~/$CONDA_ROOT/bin/conda create -y -q -n {project_name} psutil execnet
    fi

    source ~/$CONDA_ROOT/bin/activate {project_name}
    pip install automan

    cd {project_name}
    if [ -f "requirements.txt" ] ; then
        pip install -r requirements.txt
    fi
    """)

    UPDATE = dedent("""\
    #!/bin/bash

    set -e
    CONDA_ROOT={conda_root}
    ENV_FILE="{project_name}/environments.yml"
    if [ -f $ENV_FILE ] ; then
        ~/$CONDA_ROOT/bin/conda env update -q -f $ENV_FILE -n {project_name}
    fi

    source ~/$CONDA_ROOT/bin/activate {project_name}

    cd {project_name}
    if [ -f "requirements.txt" ] ; then
        pip install -r requirements.txt
    fi
    """)

    def _get_bootstrap_code(self):
        return self.BOOTSTRAP.format(
            project_name=self.project_name, conda_root=self.CONDA_ROOT
        )

    def _get_python(self, host, home):
        return os.path.join(
            home, self.CONDA_ROOT,
            'envs/{project_name}/bin/python'.format(
                project_name=self.project_name
            )
        )

    def _get_update_code(self):
        return self.UPDATE.format(
            project_name=self.project_name, conda_root=self.CONDA_ROOT
        )

    def _get_helper_scripts(self):
        """Return a space separated string of script files that you need copied over to
        the remote host.

        When overriding this, you can return None or '' if you do not need any.

        """
        return ''



# Standard libraray imports
from __future__ import print_function

from collections import deque
import json
import multiprocessing
import os
import shlex
import shutil
import subprocess
import sys
import time

# External module imports.
import psutil


def _make_command_list(command):
    if not isinstance(command, (list, tuple)):
        return shlex.split(command)
    else:
        return command


def free_cores():
    free = (1.0 - psutil.cpu_percent(interval=0.5)/100.)
    ncore = free*psutil.cpu_count(logical=False)
    return round(ncore, 0)


def total_cores():
    return psutil.cpu_count(logical=False)


def cores_required(n_core):
    if n_core < 0:
        return int(total_cores()/(-n_core))
    else:
        return n_core


def threads_required(n_thread, n_core):
    if n_thread < 0:
        return int(cores_required(n_core)*(-n_thread))
    else:
        return n_thread


class Job(object):
    def __init__(self, command, output_dir, n_core=1, n_thread=1, env=None):
        """Constructor

        Note that `n_core` is used to schedule a task on a machine which has
        that many free cores. This is not used to run the job but only used by
        the scheduler. The number can be any integer. When the number is
        negative, it will use the value of `total_cores()/(-n_core)`. This
        value may be used to set the number of threads as discussed below.

        `n_thread` is used to set the `OMP_NUM_THREADS`. Note that if
        `n_thread` is set to `None`, the environment variable is not set. If a
        positive integer is given that specific number is used. If the number
        is negative, then the number of threads is set to `n_core*(-n_thread)`,
        i.e. the product of the number of cores and the negative of the number
        given.

        """
        self.command = _make_command_list(command)
        self._given_env = env
        self.env = dict(os.environ)
        if env is not None:
            self.env.update(env)
        if n_thread is not None:
            nt = threads_required(n_thread, n_core)
            self.env['OMP_NUM_THREADS'] = str(nt)
        self.n_core = n_core
        self.n_thread = n_thread
        self.output_dir = output_dir
        self.output_already_exists = os.path.exists(self.output_dir)
        self.stderr = os.path.join(self.output_dir, 'stderr.txt')
        self.stdout = os.path.join(self.output_dir, 'stdout.txt')
        self._info_file = os.path.join(self.output_dir, 'job_info.json')
        self.proc = None

    def substitute_in_command(self, basename, substitute):
        """Replace occurrence of given basename with the substitute.

        This is useful where the user asks to run ['python', 'script.py'] and
        we wish to change the 'python' to a specific Python. Normally this is
        not needed as the PATH is set to pick up the right Python. However, in
        the rare cases where this rewriting is needed, this method is
        available.

        """
        args = []
        for arg in self.command:
            if os.path.basename(arg) == basename:
                args.append(substitute)
            else:
                args.append(arg)
        self.command = args

    def to_dict(self):
        state = dict()
        for key in ('command', 'output_dir', 'n_core', 'n_thread'):
            state[key] = getattr(self, key)
        state['env'] = self._given_env
        return state

    def pretty_command(self):
        return ' '.join(self.command)

    def get_stderr(self):
        with open(self.stderr) as fp:
            return fp.read()

    def get_stdout(self):
        with open(self.stdout) as fp:
            return fp.read()

    def get_info(self):
        return self._read_info()

    def _write_info(self, info):
        with open(self._info_file, 'w') as fp:
            json.dump(info, fp)

    def _read_info(self):
        if not os.path.exists(self._info_file):
            return {'status': 'not started'}
        with open(self._info_file, 'r') as fp:
            try:
                return json.load(fp)
            except ValueError:
                return {'status': 'running'}

    def _run(self):  # pragma: no cover
        # This is run in a multiprocessing.Process instance so does not
        # get covered.
        stdout = open(self.stdout, 'wb')
        stderr = open(self.stderr, 'wb')

        proc = subprocess.Popen(
            self.command, stdout=stdout, stderr=stderr, env=self.env
        )

        info = dict(
            start=time.ctime(), end='', status='running',
            exitcode=None, pid=proc.pid
        )
        self._write_info(info)

        proc.wait()
        status = 'error' if proc.returncode != 0 else 'done'
        info.update(end=time.ctime(), status=status, exitcode=proc.returncode)
        self._write_info(info)
        stdout.close()
        stderr.close()

    def run(self):
        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)
        self._write_info(dict(status='running', pid=None))
        self.proc = multiprocessing.Process(
            target=self._run
        )
        self.proc.start()

    def join(self):
        self.proc.join()

    def status(self):
        info = self._read_info()
        if self.proc is None and info.get('status') == 'running':
            # Either the process creating the job or the job itself
            # was killed.
            pid = info.get('pid')
            if pid is not None:
                try:
                    proc = psutil.Process(pid)
                    if not proc.is_running():
                        return 'error'
                except psutil.NoSuchProcess:
                    return 'error'
        elif self.proc is not None and info.get('status') != 'running':
            if not self.proc.is_alive():
                self.join()
                self.proc = None
        return info.get('status')

    def clean(self, force=False):
        if self.output_already_exists and not force:
            if os.path.exists(self.stdout):
                os.remove(self.stdout)
                os.remove(self.stderr)
        elif os.path.exists(self.output_dir):
            shutil.rmtree(self.output_dir)


############################################
# This class is meant to be used by execnet alone.
class _RemoteManager(object):  # pragma: no cover
    # This is run via execnet so coverage does not catch these.
    # This is used by the RemoteWorker and that is tested, so we should
    # be safe not explicitly covering this.
    def __init__(self):
        self.jobs = dict()
        self.job_count = 0
        self._setup_path()

    def _setup_path(self):
        py_dir = os.path.dirname(sys.executable)
        env_path = os.environ.get('PATH').split(os.pathsep)
        if py_dir not in env_path:
            env_path.insert(0, py_dir)
            os.environ['PATH'] = os.pathsep.join(env_path)

    def run(self, job_data):
        job = Job(**job_data)
        job.run()
        ret_val = self.job_count
        self.jobs[ret_val] = job
        self.job_count += 1
        return ret_val

    def status(self, job_id):
        if job_id in self.jobs:
            return self.jobs[job_id].status()
        else:
            return 'invalid job id %d' % job_id

    def clean(self, job_id, force=False):
        if job_id in self.jobs:
            return self.jobs[job_id].clean(force)
        else:
            return 'invalid job id %d' % job_id

    def get_stdout(self, job_id):
        return self.jobs[job_id].get_stdout()

    def get_stderr(self, job_id):
        return self.jobs[job_id].get_stderr()

    def get_info(self, job_id):
        return self.jobs[job_id].get_info()


def serve(channel):  # pragma: no cover
    """Serve the remote manager via execnet.
    """
    manager = _RemoteManager()
    while True:
        msg, data = channel.receive()
        if msg == 'free_cores':
            channel.send(free_cores())
        elif msg == 'total_cores':
            channel.send(total_cores())
        else:
            channel.send(getattr(manager, msg)(*data))
############################################


class Worker(object):
    def __init__(self):
        self.jobs = dict()
        self.running_jobs = set()
        self._total_cores = None

    def _check_running_jobs(self):
        for i in self.running_jobs.copy():
            self.status(i)

    def free_cores(self):
        return free_cores()

    def cores_required(self, n_core):
        if n_core < 0:
            return int(self.total_cores()/(-n_core))
        else:
            return n_core

    def total_cores(self):
        if self._total_cores is None:
            self._total_cores = total_cores()
        return self._total_cores

    def can_run(self, req_core):
        """Returns True if the worker can run a job with the required cores.
        """
        n_core = self.cores_required(req_core)
        if n_core == 0:
            return True
        free = self.free_cores()
        result = False
        if free >= n_core:
            self._check_running_jobs()
            jobs = self.jobs
            n_cores_used = sum(
                [self.cores_required(jobs[i].n_core)
                 for i in self.running_jobs]
            )
            if (self.total_cores() - n_cores_used) >= n_core:
                result = True
        return result

    def run(self, job):
        """Runs the job and returns a JobProxy for the job."""
        raise NotImplementedError()

    def status(self, job_id):
        """Returns status of the job."""
        raise NotImplementedError()

    def copy_output(self, job_id, dest):
        raise NotImplementedError()

    def clean(self, job_id, force=False):
        raise NotImplementedError()

    def get_stdout(self, job_id):
        raise NotImplementedError()

    def get_stderr(self, job_id):
        raise NotImplementedError()

    def get_info(self, job_id):
        raise NotImplementedError()


class JobProxy(object):
    def __init__(self, worker, job_id, job):
        self.worker = worker
        self.job_id = job_id
        self.job = job

    def free_cores(self):
        return self.worker.free_cores()

    def total_cores(self):
        return self.worker.total_cores()

    def run(self):
        print("JobProxy cannot be run")

    def status(self):
        return self.worker.status(self.job_id)

    def copy_output(self, dest):
        return self.worker.copy_output(self.job_id, dest)

    def clean(self, force=False):
        return self.worker.clean(self.job_id, force)

    def get_stdout(self):
        return self.worker.get_stdout(self.job_id)

    def get_stderr(self):
        return self.worker.get_stderr(self.job_id)

    def get_info(self):
        return self.worker.get_info(self.job_id)


class LocalWorker(Worker):
    def __init__(self):
        super(LocalWorker, self).__init__()
        self.host = 'localhost'
        self.job_count = 0

    def get_config(self):
        return dict(host='localhost')

    def run(self, job):
        count = self.job_count
        print("Running %s" % job.pretty_command())
        self.jobs[count] = job
        self.running_jobs.add(count)
        job.run()
        self.job_count += 1
        return JobProxy(self, count, job)

    def status(self, job_id):
        s = self.jobs[job_id].status()
        rj = self.running_jobs
        if s != 'running':
            rj.discard(job_id)
        return s

    def copy_output(self, job_id, dest):
        return

    def clean(self, job_id, force=False):
        if force:
            self.jobs[job_id].clean(force)

    def get_stdout(self, job_id):
        return self.jobs[job_id].get_stdout()

    def get_stderr(self, job_id):
        return self.jobs[job_id].get_stderr()

    def get_info(self, job_id):
        return self.jobs[job_id].get_info()


class RemoteWorker(Worker):
    def __init__(self, host, python, chdir=None, testing=False,
                 nfs=False):
        super(RemoteWorker, self).__init__()
        self.host = host
        self.python = python
        self.chdir = chdir
        self.testing = testing
        self.nfs = nfs
        if testing:
            spec = 'popen//python={python}'.format(python=python)
        else:
            spec = 'ssh={host}//python={python}'.format(
                host=host, python=python
            )
        if chdir is not None:
            spec += '//chdir={chdir}'.format(chdir=chdir)

        import execnet
        self.gw = execnet.makegateway(spec)
        self.channel = self.gw.remote_exec(
            "from automan import jobs; jobs.serve(channel)"
        )

    def get_config(self):
        return dict(host=self.host, python=self.python, chdir=self.chdir)

    def _call_remote(self, method, *data):
        ch = self.channel
        ch.send((method, data))
        return ch.receive()

    def free_cores(self):
        return self._call_remote('free_cores', None)

    def total_cores(self):
        if self._total_cores is None:
            self._total_cores = self._call_remote('total_cores', None)
        return self._total_cores

    def run(self, job):
        print("Running %s" % job.pretty_command())
        job_id = self._call_remote('run', job.to_dict())
        self.jobs[job_id] = job
        self.running_jobs.add(job_id)
        return JobProxy(self, job_id, job)

    def status(self, job_id):
        s = self._call_remote('status', job_id)
        rj = self.running_jobs
        if s != 'running':
            rj.discard(job_id)
        return s

    def copy_output(self, job_id, dest):
        job = self.jobs[job_id]
        if self.testing:
            src = os.path.join(self.chdir, job.output_dir)
            real_dest = os.path.join(dest, job.output_dir)
            args = [
                sys.executable, '-c',
                'import sys,shutil; shutil.copytree(sys.argv[1], sys.argv[2])',
                src, real_dest
            ]
        elif not self.nfs:
            src = '{host}:{path}'.format(
                host=self.host, path=os.path.join(self.chdir, job.output_dir)
            )
            real_dest = os.path.join(dest, os.path.dirname(job.output_dir))
            args = ['scp', '-qr', src, real_dest]
        else:
            args = []

        if args:
            print("\n" + " ".join(args))
            proc = subprocess.Popen(args)
            return proc
        else:
            return

    def clean(self, job_id, force=False):
        return self._call_remote('clean', job_id, force)

    def get_stdout(self, job_id):
        return self._call_remote('get_stdout', job_id)

    def get_stderr(self, job_id):
        return self._call_remote('get_stderr', job_id)

    def get_info(self, job_id):
        return self._call_remote('get_info', job_id)


class Scheduler(object):
    def __init__(self, root='.', worker_config=(), wait=5):
        self.workers = deque()
        self.worker_config = list(worker_config)
        self.root = os.path.abspath(os.path.expanduser(root))
        self.wait = wait
        self._completed_jobs = []
        self.jobs = []

    def _create_worker(self):
        conf = self.worker_config[len(self.workers)]
        host = conf.get('host')
        print("Starting worker on %s." % host)
        if host == 'localhost':
            w = LocalWorker()
        else:
            w = RemoteWorker(**conf)
        self.workers.append(w)
        return w

    def _get_active_workers(self):
        completed = []
        workers = set()
        for job in self.jobs:
            if job.status() in ['error', 'done']:
                completed.append(job)
            else:
                workers.add(job.worker.host)

        for job in completed:
            self.jobs.remove(job)
            self._completed_jobs.append(job)

        return workers

    def _rotate_existing_workers(self):
        worker = self.workers[0]
        self.workers.rotate(-1)
        return worker

    def _get_worker(self, n_core):
        n_configs = len(self.worker_config)
        n_running = len(self.workers)
        if n_running == n_configs:
            worker = self._rotate_existing_workers()
        else:
            active_workers = self._get_active_workers()
            if n_running > len(active_workers):
                for w in self.workers:
                    if (w.host not in active_workers) and \
                       w.can_run(n_core):
                        worker = w
                        break
                else:
                    worker = self._create_worker()
            else:
                worker = self._create_worker()
        return worker

    def save(self, fname):
        config = dict(root=self.root)
        config['workers'] = self.worker_config
        with open(fname, 'w') as fp:
            json.dump(config, fp, indent=2)

    def load(self, fname):
        with open(fname) as fp:
            config = json.load(fp)
        self.root = config.get('root')
        self.worker_config = config.get('workers')

    def add_worker(self, conf):
        self.worker_config.append(conf)

    def submit(self, job):
        proxy = None
        slept = False
        while proxy is None:
            for i in range(len(self.worker_config)):
                worker = self._get_worker(job.n_core)
                if worker.can_run(job.n_core):
                    if slept:
                        print()
                        slept = False
                    print("Job run by %s" % worker.host)
                    proxy = worker.run(job)
                    self.jobs.append(proxy)
                    break
            else:
                time.sleep(self.wait)
                slept = True
                print("\rWaiting for free worker ...", end='')
                sys.stdout.flush()
        return proxy



import psutil
import multiprocessing
import shutil
import sys
import os
import tempfile
import time
import unittest
try:
    from unittest import mock
except ImportError:
    import mock

try:
    from automan import jobs
except ImportError:
    raise unittest.SkipTest('test_jobs requires psutil')


def safe_rmtree(*args, **kw):
    if sys.platform.startswith('win'):
        try:
            shutil.rmtree(*args, **kw)
        except WindowsError:
            pass
    else:
        try:
            shutil.rmtree(*args, **kw)
        except OSError:
            pass


def test_cores_required():
    with mock.patch('automan.jobs.total_cores', return_value=4.0):
        assert jobs.cores_required(0) == 0
        assert jobs.cores_required(1) == 1
        assert jobs.cores_required(3) == 3
        assert jobs.cores_required(-1) == 4
        assert jobs.cores_required(-2) == 2
        assert jobs.cores_required(-4) == 1


def test_threads_required():
    with mock.patch('automan.jobs.total_cores', return_value=4.0):
        # n_thread, n_core
        assert jobs.threads_required(1, 1) == 1
        assert jobs.threads_required(2, 2) == 2
        assert jobs.threads_required(2, -1) == 2
        assert jobs.threads_required(-1, -1) == 4
        assert jobs.threads_required(-2, -1) == 8
        assert jobs.threads_required(-2, -2) == 4
        assert jobs.threads_required(-4, -1) == 16


class TestJob(unittest.TestCase):
    def setUp(self):
        self.root = tempfile.mkdtemp()

    def tearDown(self):
        if os.path.exists(self.root):
            safe_rmtree(self.root)

    def test_job_can_handle_string_command(self):
        # Given
        command = '''\
        python -c 'import sys;sys.stdout.write("1");sys.stderr.write("2")' '''
        j = jobs.Job(command=command, output_dir=self.root)

        # When
        j.run()
        j.join()

        # Then.
        self.assertTrue(isinstance(j.command, list))
        self.assertTrue(j.status(), 'done')
        self.assertEqual(j.get_stdout(), '1')
        self.assertEqual(j.get_stderr(), '2')

    def test_simple_job(self):
        # Given
        command = ['python', '-c',
                   'import sys;sys.stdout.write("1");sys.stderr.write("2")']
        j = jobs.Job(command=command, output_dir=self.root)

        # When
        j.run()
        j.join()

        # Then
        self.assertEqual(j.status(), 'done')
        self.assertEqual(j.output_dir, self.root)
        self.assertEqual(j.n_core, 1)
        self.assertEqual(j.n_thread, 1)
        self.assertEqual(j.get_stdout(), '1')
        self.assertEqual(j.get_stderr(), '2')
        state = j.to_dict()
        expect = dict(
            command=command, output_dir=self.root, n_core=1,
            n_thread=1, env=None
        )
        expect['command'][0] = sys.executable
        self.assertDictEqual(state, expect)

    def test_job_substitute_in_command(self):
        # Given
        j = jobs.Job(command=['python', '-c',  'print(123)'],
                     output_dir=self.root)

        # When
        sub = '/usr/bin/python'
        j.substitute_in_command('python', sub)

        # Then
        self.assertEqual(j.command[0], sub)

    def test_job_status(self):
        # Given/When
        j = jobs.Job(
            [sys.executable, '-c', 'import time; time.sleep(0.25)'],
            output_dir=self.root
        )

        # Then
        self.assertEqual(j.status(), 'not started')

        # When
        j.run()

        # Then
        self.assertEqual(j.status(), 'running')

        # When
        j.join()
        self.assertEqual(j.status(), 'done')

        # Given
        j = jobs.Job(
            [sys.executable, '-c', 'asdf'],
            output_dir=self.root
        )
        # When
        j.run()
        j.join()

        # Then
        self.assertEqual(j.status(), 'error')
        self.assertTrue('NameError' in j.get_stderr())

    def test_proc_reset_when_job_done(self):
        # Given/When
        j = jobs.Job(
            [sys.executable, '-c', 'import time; time.sleep(0.25)'],
            output_dir=self.root
        )

        # Then
        self.assertEqual(j.status(), 'not started')
        self.assertIsNone(j.proc)

        # When
        j.run()

        # Then
        self.assertEqual(j.status(), 'running')
        self.assertIsNotNone(j.proc)

        # When
        j.join()
        self.assertEqual(j.status(), 'done')
        self.assertIsNone(j.proc)

    def test_reset_proc_when_job_status_error(self):
        j = jobs.Job(
            [sys.executable, '--junk'],
            output_dir=self.root,
            n_thread=4,
        )

        # Then
        self.assertEqual(j.status(), 'not started')
        self.assertIsNone(j.proc)

        # When
        j.run()

        self.assertEqual(j.status(), 'running')
        self.assertIsNotNone(j.proc)

        # When
        j.join()

        # Then
        self.assertEqual(j.status(), 'error')
        self.assertIsNone(j.proc)

    def test_that_job_sets_env_var(self):
        # Given/When
        j = jobs.Job(
            [sys.executable, '-c', 'import os;print(os.environ.get("FOO"))'],
            output_dir=self.root,
            env=dict(FOO='hello')
        )
        j.run()
        j.join()

        # Then
        print(j.get_stdout(), j.get_stderr())
        self.assertEqual(j.status(), 'done')
        self.assertEqual(j.get_stdout().strip(), 'hello')

    @mock.patch('automan.jobs.total_cores', return_value=2.0)
    def test_that_job_sets_omp_var(self, mock_total_cores):
        j = jobs.Job(
            [sys.executable, '-c',
             'import os;print(os.environ.get("OMP_NUM_THREADS"))'],
            output_dir=self.root,
            n_thread=4,
        )
        j.run()
        j.join()

        # Then
        self.assertEqual(j.status(), 'done')
        self.assertEqual(j.get_stdout().strip(), '4')
        self.assertEqual(j.env.get('OMP_NUM_THREADS'), '4')

        # When
        j = jobs.Job(
            [sys.executable, '-c', 'print(1)'],
            output_dir=self.root,
            n_thread=None,
        )

        # Then
        self.assertFalse('OMP_NUM_THREADS' in j.env)

        # When
        j = jobs.Job(
            [sys.executable, '-c', 'print(1)'],
            output_dir=self.root,
            n_thread=-2, n_core=1
        )

        # Then
        self.assertEqual(j.env.get('OMP_NUM_THREADS'), '2')

        # When
        j = jobs.Job(
            [sys.executable, '-c', 'print(1)'],
            output_dir=self.root,
            n_thread=-2, n_core=-1
        )

        # Then
        self.assertEqual(j.env.get('OMP_NUM_THREADS'), '4')

    def test_free_cores(self):
        n = jobs.free_cores()
        self.assertTrue(n >= 0)
        self.assertTrue(n <= multiprocessing.cpu_count())

    def test_total_cores(self):
        n = jobs.total_cores()
        self.assertTrue(n >= 0)
        self.assertTrue(n <= multiprocessing.cpu_count())

    def test_status_when_job_is_incorrect(self):
        j = jobs.Job(
            [sys.executable, '--junk'],
            output_dir=self.root,
            n_thread=4,
        )
        j.run()
        j.join()

        # Then
        self.assertEqual(j.status(), 'error')
        info = j.get_info()
        self.assertEqual(info['status'], 'error')
        self.assertTrue(info['exitcode'] != 0)

        # Now retry and the status should be the same.
        j1 = jobs.Job(
            [sys.executable, '--junk'],
            output_dir=self.root,
            n_thread=4,
        )

        self.assertEqual(j1.status(), 'error')

    def test_status_when_job_is_rerun(self):
        # Given
        command = ['python', '-c', 'print(123)']
        j = jobs.Job(command=command, output_dir=self.root)
        j.run()
        j.join()

        # When
        j1 = jobs.Job(command=command, output_dir=self.root)

        # Then
        self.assertEqual(j1.status(), 'done')

    def test_clean_removes_new_output_directory(self):
        # Given
        out_dir = os.path.join(self.root, 'junk')
        command = ['python', '-c', 'print(123)']
        j = jobs.Job(command=command, output_dir=out_dir)
        j.run()
        j.join()

        # When
        j.clean()

        # Then
        self.assertFalse(os.path.exists(out_dir))

    def test_clean_does_not_remove_existing_output_directory(self):
        # Given
        command = ['python', '-c', 'print(123)']
        j = jobs.Job(command=command, output_dir=self.root)
        j.run()
        j.join()

        # When
        j.clean()

        # Then
        self.assertFalse(os.path.exists(j.stdout))
        self.assertFalse(os.path.exists(j.stderr))
        self.assertTrue(os.path.exists(self.root))

    def test_job_status_when_process_killed(self):
        # Given
        cmd = ['python', '-c', 'import time; time.sleep(100)']
        j = jobs.Job(command=cmd, output_dir=self.root)
        j.run()
        wait_until(lambda: j.get_info()['pid'] is None)

        # When

        # Kill the Python process the job is running, as well as its subprocess.
        pid = j.get_info()['pid']
        proc = psutil.Process(pid)
        j.proc.kill()
        proc.kill()

        def _check_proc():
            try:
                proc.status()
                return True
            except psutil.NoSuchProcess:
                return False

        wait_until(_check_proc)

        j = jobs.Job(command=cmd, output_dir=self.root)

        # Then
        self.assertEqual(j.status(), 'error')


def wait_until(cond, timeout=1, wait=0.1):
    t = 0.0
    while cond():
        time.sleep(wait)
        t += wait
        if t > timeout:
            break


class TestLocalWorker(unittest.TestCase):
    def setUp(self):
        self.root = tempfile.mkdtemp()

    def tearDown(self):
        safe_rmtree(self.root)

    @mock.patch('automan.jobs.total_cores', return_value=4.0)
    @mock.patch('automan.jobs.free_cores', return_value=2.0)
    def test_worker_computes_correct_cores_required(
            self, mock_free_cores, mock_total_cores
    ):
        # Given
        w = jobs.Worker()
        self.assertEqual(w.total_cores(), 4.0)
        self.assertEqual(w.free_cores(), 2.0)

        # When/Then
        self.assertEqual(w.cores_required(1), 1)
        self.assertEqual(w.cores_required(2), 2)
        self.assertEqual(w.cores_required(0), 0)
        self.assertEqual(w.cores_required(-1), 4)
        self.assertEqual(w.cores_required(-2), 2)

        self.assertEqual(w.can_run(0), True)
        self.assertEqual(w.can_run(1), True)
        self.assertEqual(w.can_run(2), True)
        self.assertEqual(w.can_run(-2), True)
        self.assertEqual(w.can_run(-1), False)

    @mock.patch('automan.jobs.free_cores', return_value=2.0)
    def test_scheduler_works_with_local_worker(self, mock_free_cores):
        # Given
        s = jobs.Scheduler(worker_config=[dict(host='localhost')])

        # When
        j = jobs.Job(
            [sys.executable, '-c', 'import time; time.sleep(0.05); print(1)'],
            output_dir=self.root
        )
        proxy = s.submit(j)

        # Then
        wait_until(lambda: proxy.status() != 'done', timeout=2)
        self.assertEqual(proxy.status(), 'done')
        self.assertEqual(proxy.get_stderr(), '')
        self.assertEqual(proxy.get_stdout().strip(), '1')
        info = proxy.get_info()
        self.assertEqual(info['status'], 'done')
        self.assertEqual(info['exitcode'], 0)


class TestRemoteWorker(unittest.TestCase):
    def setUp(self):
        self.root = tempfile.mkdtemp()
        try:
            import execnet
        except ImportError:
            raise unittest.SkipTest('This test requires execnet')

    def tearDown(self):
        safe_rmtree(self.root)

    def test_free_cores(self):
        # Given
        r = jobs.RemoteWorker(
            host='localhost', python=sys.executable, testing=True
        )
        # Then.
        n = r.free_cores()
        self.assertTrue(n >= 0)
        self.assertTrue(n <= multiprocessing.cpu_count())

    def test_simple(self):
        # Given
        r = jobs.RemoteWorker(
            host='localhost', python=sys.executable, testing=True
        )

        # When
        j = jobs.Job(
            [sys.executable, '-c', 'import time; time.sleep(0.05); print(1)'],
            output_dir=self.root
        )
        proxy = r.run(j)

        # Then
        wait_until(lambda: proxy.status() != 'done')
        self.assertEqual(proxy.status(), 'done')
        self.assertEqual(proxy.get_stderr(), '')
        self.assertEqual(proxy.get_stdout().strip(), '1')
        info = proxy.get_info()
        self.assertEqual(info['status'], 'done')
        self.assertEqual(info['exitcode'], 0)

    def test_remote_worker_does_not_copy_when_nfs_is_set(self):
        # Given
        r = jobs.RemoteWorker(
            host='localhost', python=sys.executable, testing=True, nfs=True
        )

        # When
        j = jobs.Job(
            [sys.executable, '-c', 'import time; time.sleep(0.05); print(1)'],
            output_dir=self.root
        )
        proxy = r.run(j)
        wait_until(lambda: proxy.status() != 'done')

        # Now set testing to False and check the copying
        r.testing = False
        ret = r.copy_output(proxy.job_id, '.')

        # Then
        self.assertEqual(ret, None)


class TestScheduler(unittest.TestCase):
    def setUp(self):
        self.root = tempfile.mkdtemp()
        self.count = 0
        try:
            import execnet
        except ImportError:
            raise unittest.SkipTest('This test requires execnet')

    def tearDown(self):
        safe_rmtree(self.root)

    def _make_dummy_job(self, n_core=1, sleep=0.05):
        output = os.path.join(self.root, 'job%d' % self.count)
        job = jobs.Job(
            [sys.executable, '-c',
             'import time; time.sleep(%f); print(1)' % sleep],
            output_dir=output,
            n_core=n_core
        )
        self.count += 1
        return job

    @mock.patch('automan.jobs.LocalWorker')
    def test_scheduler_does_not_start_worker_when_created(self, mock_lw):
        # Given
        config = [dict(host='localhost')]

        # When
        s = jobs.Scheduler(worker_config=config)

        # Then
        self.assertEqual(mock_lw.call_count, 0)
        self.assertEqual(len(s.workers), 0)

    @mock.patch('automan.jobs.LocalWorker')
    def test_scheduler_starts_worker_on_submit(self, mock_lw):
        attrs = {'host': 'localhost', 'free_cores.return_value': 2}
        mock_lw.return_value = mock.MagicMock(**attrs)

        # Given
        config = [dict(host='localhost')]
        s = jobs.Scheduler(worker_config=config)
        j = jobs.Job(
            [sys.executable, '-c', 'print(1)'],
            output_dir=self.root
        )

        # When
        s.submit(j)

        # Then
        self.assertEqual(mock_lw.call_count, 1)
        self.assertEqual(len(s.workers), 1)

    @mock.patch.object(jobs.RemoteWorker, 'free_cores', return_value=2.0)
    def test_scheduler_only_creates_required_workers(self, mock_remote_worker):
        # Given
        config = [
            dict(host='host1', python=sys.executable, testing=True),
            dict(host='host2', python=sys.executable, testing=True),
        ]
        s = jobs.Scheduler(worker_config=config)
        j = self._make_dummy_job()

        # When
        proxy = s.submit(j)

        # Then
        self.assertEqual(len(s.workers), 1)
        self.assertEqual(proxy.worker.host, 'host1')

        # Wait for this job to end and then see what happens
        # When a new job is submitted.
        self._wait_while_not_done(proxy, 25)

        # When
        j = self._make_dummy_job()
        s.submit(j)

        # Then
        self.assertEqual(len(s.workers), 1)
        self.assertEqual(proxy.worker.host, 'host1')

        # Running two jobs in a row should produce two workers.
        # When
        j = self._make_dummy_job()
        proxy = s.submit(j)

        # Then
        self.assertEqual(len(s.workers), 2)
        self.assertEqual(proxy.worker.host, 'host2')

        # Adding more should work.

        # When
        j = self._make_dummy_job()
        proxy = s.submit(j)
        j = self._make_dummy_job()
        proxy1 = s.submit(j)

        # Then
        self.assertEqual(len(s.workers), 2)
        self._wait_while_not_done(proxy, 15)
        self._wait_while_not_done(proxy1, 15)

        self.assertEqual(proxy.status(), 'done')
        self.assertEqual(proxy.worker.host, 'host1')
        self.assertEqual(proxy1.worker.host, 'host2')

    @mock.patch('automan.jobs.total_cores', return_value=2.0)
    @mock.patch('automan.jobs.free_cores', return_value=2.0)
    def test_scheduler_should_not_overload_worker(self, m_total_cores,
                                                  m_free_cores):
        # Given
        n_core = jobs.total_cores()
        config = [dict(host='localhost')]
        s = jobs.Scheduler(worker_config=config, wait=0.5)

        j1 = self._make_dummy_job(n_core, sleep=0.5)
        j2 = self._make_dummy_job(n_core, sleep=0.5)
        j3 = self._make_dummy_job(n_core, sleep=0.5)
        j4 = self._make_dummy_job(0, sleep=0.5)

        # When
        proxy1 = s.submit(j1)
        proxy2 = s.submit(j2)
        proxy3 = s.submit(j3)
        proxy4 = s.submit(j4)

        # Then
        self.assertEqual(len(s.workers), 1)
        # Basically, submit will wait for the existing jobs to complete.
        # Therefore when s.submit(j3) is called it should wait until a
        # free worker is available.
        self.assertEqual(proxy1.status(), 'done')
        self.assertEqual(proxy2.status(), 'done')
        self.assertEqual(proxy3.status(), 'running')

        # Proxy4 will be running since it needs no cores.
        self.assertEqual(proxy4.status(), 'running')
        self._wait_while_not_done(proxy3, 20)
        self.assertEqual(proxy3.status(), 'done')
        self._wait_while_not_done(proxy4, 15)
        self.assertEqual(proxy4.status(), 'done')

    def _wait_while_not_done(self, proxy, n_count, sleep=0.1):
        count = 0
        while proxy.status() != 'done' and count < n_count:
            time.sleep(sleep)
            count += 1



import argparse
import os

import numpy as np


def compute_powers(r_max, power):
    """Compute the powers of the integers upto r_max and return the result.
    """
    x = np.arange(0, r_max + 1)
    y = np.power(x, power)
    return x, y


def main():
    p = argparse.ArgumentParser()
    p.add_argument(
        '--power', type=float, default=2.0,
        help='Power to calculate'
    )
    p.add_argument(
        '--max', type=int, default=10,
        help='Maximum integer that we must raise to the given power'
    )
    p.add_argument(
        '--output-dir', type=str, default='.',
        help='Output directory to generate file.'
    )
    opts = p.parse_args()

    x, y = compute_powers(opts.max, opts.power)

    fname = os.path.join(opts.output_dir, 'results.npz')
    np.savez(fname, x=x, y=y)


if __name__ == '__main__':
    main()



"""Utility functions for automation scripts.
"""
import collections
import itertools as IT


def dprod(a, b):
    '''Multiplies the given list of dictionaries `a` and `b`.

    This makes a list of new dictionaries which is the product of the given
    two dictionaries.

    **Example**

    >>> dprod(mdict(a=[1, 2], b=['xy']), mdict(c='ab'))
    [{'a': 1, 'b': 'xy', 'c': 'a'},
     {'a': 1, 'b': 'xy', 'c': 'b'},
     {'a': 2, 'b': 'xy', 'c': 'a'},
     {'a': 2, 'b': 'xy', 'c': 'b'}]

    '''
    return [
        dict(IT.chain(x.items(),  y.items())) for x, y in IT.product(a, b)
    ]


def styles(sims):
    """Cycles over a set of possible styles to use for plotting.

    The method is passed a sequence of the Simulation instances. This should
    return an iterator which produces a dictionary each time containing a set
    of keyword arguments to be used for a particular plot.

    **Parameters**

    sims: sequence
        Sequence of `Simulation` objects.

    **Returns**

    An iterator which produces a dictionary containing a set of kwargs to be
    used for the plotting. Can also return an iterable containing
    dictionaries.

    """
    ls = [dict(color=x[0], linestyle=x[1]) for x in
          IT.product("kbgr", ["-", "--", "-.", ":"])]
    return IT.cycle(ls)


def compare_runs(sims, method, labels, exact=None, styles=styles):
    """Given a sequence of Simulation instances, a method name, the labels to
    compare and an optional method name for an exact solution, this calls the
    methods with the appropriate parameters for each simulation.

    **Parameters**

    sims: sequence
        Sequence of `Simulation` objects.
    method: str or callable
        Name of a method on each simulation method to call for plotting.
        Or a callable which is passed the simulation instance and any kwargs.
    labels: sequence
        Sequence of parameters to use as labels for the plot.
    exact: str or callable
        Name of a method that produces an exact solution plot
        or a callable that will be called.
    styles: callable: returns an iterator/iterable of style keyword arguments.
        Defaults to the ``styles`` function defined in this module.
    """
    ls = styles(sims)
    if isinstance(ls, collections.abc.Iterable):
        ls = iter(ls)
    if exact is not None:
        if isinstance(exact, str):
            getattr(sims[0], exact)(**next(ls))
        else:
            exact(sims[0], **next(ls))
    for s in sims:
        if isinstance(method, str):
            m = getattr(s, method)
            m(label=s.get_labels(labels), **next(ls))
        else:
            method(s, label=s.get_labels(labels), **next(ls))


def filter_cases(runs, predicate=None, **params):
    """Given a sequence of simulations and any additional parameters, filter
    out all the cases having exactly those parameters and return a list of
    them.

    One may also pass a callable to filter the cases using the `predicate`
    keyword argument. If this is not a callable, it is treated as a parameter.
    If `predicate` is passed though, the other keyword arguments are ignored.

    """
    if predicate is not None:
        if callable(predicate):
            return list(filter(predicate, runs))
        else:
            params['predicate'] = predicate

    def _check_match(run):
        for param, expected in params.items():
            if param not in run.params or run.params[param] != expected:
                return False
        return True

    return list(filter(_check_match, runs))


def filter_by_name(cases, names):
    """Filter a sequence of Simulations by their names.  That is, if the case
    has a name contained in the given `names`, it will be selected.
    """
    if isinstance(names, str):
        names = [names]
    return sorted(
        [x for x in cases if x.name in names],
        key=lambda x: names.index(x.name)
    )


def mdict(**kw):
    '''Expands out the passed kwargs into a list of dictionaries.

    Each kwarg value is expected to be a sequence. The resulting list of
    dictionaries is the product of the different values and the same keys.

    **Example**

    >>> mdict(a=[1, 2], b='xy')
    [{'a': 1, 'b': 'x'},
     {'a': 1, 'b': 'y'},
     {'a': 2, 'b': 'x'},
     {'a': 2, 'b': 'y'}]

    '''
    keys = list(kw.keys())
    return [dict(zip(keys, opts)) for opts in IT.product(*kw.values())]


def opts2path(opts, keys=None, ignore=None, kmap=None):
    '''Renders the given options as a path name.

    **Parameters**

    opts: dict
        dictionary of options
    keys: list
        Keys of the options use.
    ignore: list
        Ignore these keys in the options.
    kmap: dict
        map the key names through this dict.

    **Examples**

    >>> opts2path(dict(x=1, y='hello', z=0.1))
    'x_1_hello_z_0.1'
    >>> opts2path(dict(x=1, y='hello', z=0.1), keys=['x'])
    'x_1'
    >>> opts2path(dict(x=1, y='hello', z=0.1), ignore=['x'])
    'hello_z_0.1'
    >>> opts2path(dict(x=1, y='hello', z=0.1), kmap=dict(x='XX'))
    'XX_1_hello_z_0.1'
    '''
    keys = set(opts.keys()) if keys is None else set(keys)
    ignore = [] if ignore is None else ignore
    keymap = {} if kmap is None else kmap
    for x in ignore:
        keys.discard(x)
    keys = sorted(x for x in keys if x in opts)

    def _key2name(k):
        v = opts[k]
        r = keymap.get(k, '')
        if r:
            return f'{r}_{v}'
        elif isinstance(v, str):
            return v
        else:
            return f'{k}_{v}'

    return '_'.join([_key2name(k) for k in keys])



from __future__ import print_function

import json
import os
from os.path import dirname
import shutil
import sys
import tempfile
from textwrap import dedent
import unittest

try:
    from unittest import mock
except ImportError:
    import mock

from automan.jobs import Job
from automan.cluster_manager import ClusterManager
from automan.conda_cluster_manager import CondaClusterManager
from automan.edm_cluster_manager import EDMClusterManager
from .test_jobs import wait_until


ROOT_DIR = dirname(dirname(dirname(__file__)))


class MyClusterManager(ClusterManager):

    BOOTSTRAP = dedent("""\
        #!/bin/bash

        set -e
        python3 -m venv --system-site-packages envs/{project_name}
        source envs/{project_name}/bin/activate
        cd ROOT
        python setup.py install
        """)

    UPDATE = dedent("""\
         #!/bin/bash
         echo "update"
         """)

    def _get_helper_scripts(self):
        return None


class TestClusterManager(unittest.TestCase):

    def setUp(self):
        self.cwd = os.getcwd()
        self.root = tempfile.mkdtemp()
        self.proj_root = os.path.join(self.root, 'project')
        os.makedirs(self.proj_root)
        os.chdir(self.proj_root)

    def tearDown(self):
        os.chdir(self.cwd)
        if os.path.exists(self.root):
            shutil.rmtree(self.root)

    def _get_config(self):
        with open(os.path.join(self.proj_root, 'config.json'), 'r') as fp:
            data = json.load(fp)
        return data

    def test_new_config_created_on_creation(self):
        # Given/When
        ClusterManager()

        # Then
        config = self._get_config()

        self.assertEqual(config.get('root'), 'automan')
        self.assertEqual(config.get('project_name'),
                         os.path.basename(self.proj_root))
        self.assertEqual(os.path.realpath(config.get('sources')[0]),
                         os.path.realpath(self.proj_root))
        workers = config.get('workers')
        self.assertEqual(len(workers), 1)
        self.assertEqual(workers[0]['host'], 'localhost')

    @mock.patch.object(ClusterManager, '_bootstrap')
    def test_add_worker(self, mock_bootstrap):
        # Given
        cm = ClusterManager()

        # When
        cm.add_worker('host', home='/home/foo', nfs=False)

        # Then
        mock_bootstrap.assert_called_with('host', '/home/foo')
        config = self._get_config()
        workers = config.get('workers')
        self.assertEqual(len(workers), 2)
        hosts = sorted(x['host'] for x in workers)
        self.assertEqual(hosts, ['host', 'localhost'])

    @mock.patch.object(ClusterManager, '_bootstrap')
    def test_create_scheduler(self, mock_bootstrap):
        # Given
        cm = ClusterManager()
        cm.add_worker('host', home='/home/foo', nfs=False)

        # When
        s = cm.create_scheduler()

        # Then
        confs = s.worker_config
        hosts = sorted([x['host'] for x in confs])
        self.assertEqual(hosts, ['host', 'localhost'])

    @mock.patch.object(ClusterManager, '_bootstrap')
    @mock.patch.object(ClusterManager, '_update_sources')
    @mock.patch.object(ClusterManager, '_rebuild')
    def test_update(self, mock_rebuild, mock_update_sources, mock_bootstrap):
        # Given
        cm = ClusterManager()
        cm.add_worker('host', home='/home/foo', nfs=False)

        # When
        cm.update()

        # Then
        mock_update_sources.assert_called_with('host', '/home/foo')
        mock_rebuild.assert_called_with('host', '/home/foo')

    @mock.patch.object(ClusterManager, 'add_worker')
    @mock.patch.object(ClusterManager, 'update')
    def test_cli(self, mock_update, mock_add_worker):
        # Given
        cm = ClusterManager()

        # When
        cm.cli(['-a', 'host', '--home', 'home', '--nfs'])

        # Then
        mock_add_worker.assert_called_with('host', 'home', True)

    @unittest.skipIf((sys.version_info < (3, 3)) or
                     sys.platform.startswith('win'),
                     'Test requires Python 3.x and a non-Windows system.')
    def test_remote_bootstrap_and_sync(self):
        if not os.path.exists('setup.py'):
            raise unittest.SkipTest(
                'This test requires to be run from the automan source directory.'
                )

        # Given
        cm = MyClusterManager(exclude_paths=['outputs/'], testing=True)
        cm.BOOTSTRAP = cm.BOOTSTRAP.replace('ROOT', self.cwd)
        output_dir = os.path.join(self.proj_root, 'outputs')
        os.makedirs(output_dir)

        # Remove the default localhost worker.
        cm.workers = []

        # When
        cm.add_worker('host', home=self.root, nfs=False)

        # Then
        self.assertEqual(len(cm.workers), 1)
        worker = cm.workers[0]
        self.assertEqual(worker['host'], 'host')
        project_name = cm.project_name
        self.assertEqual(project_name, os.path.basename(self.proj_root))
        py = os.path.join(self.root, 'automan', 'envs', project_name,
                          'bin', 'python')
        self.assertEqual(worker['python'], py)
        chdir = os.path.join(self.root, 'automan', project_name)
        self.assertEqual(worker['chdir'], chdir)

        # Given
        cmd = ['python', '-c', 'import sys; print(sys.executable)']
        job = Job(command=cmd, output_dir=output_dir)

        s = cm.create_scheduler()

        # When
        proxy = s.submit(job)

        # Then
        wait_until(lambda: proxy.status() != 'done')

        self.assertEqual(proxy.status(), 'done')
        output = proxy.get_stdout().strip()
        self.assertEqual(os.path.realpath(output), os.path.realpath(py))

        # Test to see if updating works.

        # When
        with open(os.path.join(self.proj_root, 'script.py'), 'w') as f:
            f.write('print("hello")\n')

        cm.update()

        # Then
        dest = os.path.join(self.root, 'automan', project_name, 'script.py')
        self.assertTrue(os.path.exists(dest))

    @mock.patch.object(ClusterManager, '_bootstrap')
    @mock.patch.object(ClusterManager, '_delete_outputs')
    def test_delete_outputs_from_hosts(self, mock_delete_outputs,
                                       mock_bootstrap):
        # Given
        cm = ClusterManager()
        cm.add_worker('host', home='/home/foo', nfs=False)
        cm.add_worker('host1', home='/home/foo1', nfs=False)

        # When
        cm.delete('outputs', ['host', 'host1'])

        # Then
        mock_delete_outputs.assert_any_call('host1', '/home/foo1', 'outputs')
        mock_delete_outputs.assert_any_call('host', '/home/foo', 'outputs')

    @mock.patch.object(ClusterManager, '_bootstrap')
    @mock.patch.object(ClusterManager, '_delete_outputs')
    def test_do_not_delete_outputs_from_hosts_when_nfs_true(
            self, mock_delete_outputs, mock_bootstrap
    ):
        # Given
        cm = ClusterManager()
        cm.add_worker('host', home='/home/foo', nfs=True)
        cm.add_worker('host1', home='/home/foo1', nfs=True)

        # When
        cm.delete('outputs', ['host', 'host1'])

        # Then
        mock_delete_outputs.assert_not_called()

    @mock.patch.object(ClusterManager, '_bootstrap')
    @mock.patch.object(ClusterManager, '_delete_outputs')
    def test_delete_outputs_from_all_hosts(self, mock_delete_outputs,
                                           mock_bootstrap):
        # Given
        cm = ClusterManager()
        cm.add_worker('host', home='/home/foo', nfs=False)
        cm.add_worker('host1', home='/home/foo1', nfs=False)

        # When
        cm.delete('outputs', ['all'])

        # Then
        mock_delete_outputs.assert_any_call('host1', '/home/foo1', 'outputs')
        mock_delete_outputs.assert_any_call('host', '/home/foo', 'outputs')


class TestCondaClusterManager(unittest.TestCase):
    def setUp(self):
        self.cwd = os.getcwd()
        self.root = tempfile.mkdtemp()
        os.chdir(self.root)

    def tearDown(self):
        os.chdir(self.cwd)
        if os.path.exists(self.root):
            shutil.rmtree(self.root)

    @mock.patch("automan.conda_cluster_manager.CondaClusterManager.CONDA_ROOT",
                'TEST_ROOT')
    def test_overloaded_methods(self):
        # Given
        cm = CondaClusterManager()

        # When/Then
        python = cm._get_python('foo', 'blah')
        name = os.path.basename(self.root)
        self.assertEqual(python, os.path.join(
            'blah', 'TEST_ROOT', 'envs/{name}/bin/python'.format(name=name)
        ))

        code = cm._get_bootstrap_code()
        self.assertTrue('TEST_ROOT' in code)

        code = cm._get_update_code()
        self.assertTrue('TEST_ROOT' in code)

        self.assertEqual(cm._get_helper_scripts(), '')


class TestEDMClusterManager(unittest.TestCase):
    def setUp(self):
        self.cwd = os.getcwd()
        self.root = tempfile.mkdtemp()
        os.chdir(self.root)

    def tearDown(self):
        os.chdir(self.cwd)
        if os.path.exists(self.root):
            shutil.rmtree(self.root)

    @mock.patch("automan.edm_cluster_manager.EDMClusterManager.EDM_ROOT",
                'TEST_ROOT')
    def test_overloaded_methods(self):
        # Given
        cm = EDMClusterManager()

        # When/Then
        python = cm._get_python('foo', 'bar')
        name = os.path.basename(self.root)
        self.assertEqual(python, os.path.join(
            'bar', 'TEST_ROOT', 'envs/{name}/bin/python'.format(name=name)
        ))

        code = cm._get_bootstrap_code()
        self.assertTrue('TEST_ROOT' in code)

        code = cm._get_update_code()
        self.assertTrue('TEST_ROOT' in code)

        self.assertEqual(cm._get_helper_scripts(), '')



from __future__ import print_function

import os
import sys
import tempfile
import unittest

try:
    from unittest import mock
except ImportError:
    import mock

from automan.automation import (
    Automator, CommandTask, FileCommandTask, Problem, PySPHProblem, RunAll,
    Simulation, SolveProblem, TaskRunner
)
try:
    from automan.jobs import Scheduler, RemoteWorker
except ImportError:
    raise unittest.SkipTest('test_jobs requires psutil')

from automan.cluster_manager import ClusterManager

from automan.tests.test_jobs import wait_until, safe_rmtree


class MySimulation(Simulation):
    @property
    def data(self):
        if self._results is None:
            with open(self.input_path('results.dat')) as fp:
                self._results = fp.read()
        return self._results


class EllipticalDrop(PySPHProblem):
    """We define a simple example problem which we will run using the
    automation framework.

    In this case we run two variants of the elliptical drop problem.

    The setup method defines the cases to run which are simply Simulation
    instances.

    The get_commands returns the actual commands to run.

    The run method does the post-processing, after the simulations are done.

    """

    def get_name(self):
        return 'elliptical_drop'

    def setup(self):
        # Two cases, one with update_h and one without.
        cmd = 'python -m automan.tests.example'

        # If self.cases is set, the get_requires method will do the right
        # thing.
        self.cases = [
            MySimulation(
                root=self.input_path('update_h'),
                base_command=cmd,
                job_info=dict(n_core=1, n_thread=1),
                update_h=None
            ),
            MySimulation(
                root=self.input_path('no_update_h'),
                base_command=cmd,
                job_info=dict(n_core=1, n_thread=1),
                no_update_h=None
            ),
        ]

    def run(self):
        self.make_output_dir()
        no_update = self.cases[0].data
        update = self.cases[1].data
        output = open(self.output_path('result.txt'), 'w')
        output.write('no_update_h: %s\n' % no_update)
        output.write('update_h: %s\n' % update)
        output.close()


class TestAutomationBase(unittest.TestCase):
    def setUp(self):
        self.cwd = os.getcwd()
        self.root = tempfile.mkdtemp()
        os.chdir(self.root)
        self.sim_dir = 'sim'
        self.output_dir = 'output'
        patch = mock.patch(
            'automan.jobs.free_cores', return_value=2
        )
        patch.start()
        self.addCleanup(patch.stop)

    def tearDown(self):
        os.chdir(self.cwd)
        if os.path.exists(self.root):
            safe_rmtree(self.root)


class TestTaskRunner(TestAutomationBase):
    def _make_scheduler(self):
        worker = dict(host='localhost')
        s = Scheduler(root='.', worker_config=[worker], wait=0.1)
        return s

    def _get_time(self, path):
        with open(os.path.join(path, 'stdout.txt')) as f:
            t = float(f.read())
        return t

    @mock.patch('automan.jobs.total_cores', return_value=2)
    def test_task_runner_waits_for_tasks_in_the_end(self, m_t_cores):
        # Given
        s = self._make_scheduler()
        cmd = 'python -c "import sys, time; time.sleep(0.1); sys.exit(1)"'
        ct1_dir = os.path.join(self.sim_dir, '1')
        ct2_dir = os.path.join(self.sim_dir, '2')
        ct3_dir = os.path.join(self.sim_dir, '3')
        ct1 = CommandTask(cmd, output_dir=ct1_dir)
        ct2 = CommandTask(cmd, output_dir=ct2_dir)
        ct3 = CommandTask(cmd, output_dir=ct3_dir)

        # When
        t = TaskRunner(tasks=[ct1, ct2, ct3], scheduler=s)
        n_errors = t.run(wait=0.1)

        # Then
        # All the tasks may have been run but those that ran will fail.
        self.assertEqual(n_errors + len(t.todo), 3)
        self.assertTrue(n_errors > 0)

    @mock.patch('automan.jobs.total_cores', return_value=2)
    def test_task_runner_checks_for_error_in_running_tasks(self, m_t_cores):
        # Given
        s = self._make_scheduler()
        cmd = 'python -c "import sys, time; time.sleep(0.1); sys.exit(1)"'
        ct1_dir = os.path.join(self.sim_dir, '1')
        ct2_dir = os.path.join(self.sim_dir, '2')
        ct3_dir = os.path.join(self.sim_dir, '3')
        job_info = dict(n_core=2, n_thread=2)
        ct1 = CommandTask(cmd, output_dir=ct1_dir, job_info=job_info)
        ct2 = CommandTask(cmd, output_dir=ct2_dir, job_info=job_info)
        ct3 = CommandTask(cmd, output_dir=ct3_dir, job_info=job_info)

        # When
        t = TaskRunner(tasks=[ct1, ct2, ct3], scheduler=s)

        # Then
        self.assertEqual(len(t.todo), 3)

        # When
        n_errors = t.run(wait=0.1)

        # Then
        # In this case, two tasks should have run and one should not have run
        # as the other two had errors.
        self.assertEqual(n_errors, 2)
        self.assertEqual(len(t.todo), 1)
        self.assertTrue(os.path.exists(ct3_dir))
        self.assertTrue(os.path.exists(ct2_dir))
        self.assertFalse(os.path.exists(ct1_dir))

    @mock.patch('automan.jobs.total_cores', return_value=2)
    def test_task_runner_doesnt_block_on_problem_with_error(self, m_t_cores):
        # Given
        class A(Problem):
            def get_requires(self):
                cmd = ('python -c "import sys, time; time.sleep(0.1); '
                       'sys.exit(1)"')
                ct = CommandTask(cmd, output_dir=self.input_path())
                return [('task1', ct)]

            def run(self):
                self.make_output_dir()

        s = self._make_scheduler()

        # When
        task = RunAll(
            simulation_dir=self.sim_dir, output_dir=self.output_dir,
            problem_classes=[A]
        )
        t = TaskRunner(tasks=[task], scheduler=s)
        n_error = t.run(wait=0.1)

        # Then
        self.assertTrue(n_error > 0)

        # For the case where there is a match expression
        # When
        problem = A(self.sim_dir, self.output_dir)
        problem.clean()

        task = RunAll(
            simulation_dir=self.sim_dir, output_dir=self.output_dir,
            problem_classes=[A], match='*task1'
        )
        t = TaskRunner(tasks=[task], scheduler=s)
        n_error = t.run(wait=0.1)

        # Then
        self.assertTrue(n_error > 0)

    def test_task_runner_does_not_add_repeated_tasks(self):
        # Given
        s = self._make_scheduler()
        cmd = 'python -c "print(1)"'
        ct1 = CommandTask(cmd, output_dir=self.sim_dir)
        ct2 = CommandTask(cmd, output_dir=self.sim_dir)

        # When
        t = TaskRunner(tasks=[ct1, ct2, ct1], scheduler=s)

        # Then
        self.assertEqual(len(t.todo), 1)

    def test_problem_depending_on_other_problems(self):
        # Given
        class A(Problem):
            def get_requires(self):
                cmd = 'python -c "print(1)"'
                # Can return tasks ...
                ct = CommandTask(cmd, output_dir=self.sim_dir)
                return [('task1', ct)]

            def run(self):
                self.make_output_dir()

        class B(Problem):
            def get_requires(self):
                # or return Problem instances ...
                return [('a', A(self.sim_dir, self.out_dir))]

            def run(self):
                self.make_output_dir()

        class C(Problem):
            def get_requires(self):
                # ... or Problem subclasses
                return [('a', A), ('b', B)]

            def run(self):
                self.make_output_dir()

        s = self._make_scheduler()

        # When
        task = RunAll(
            simulation_dir=self.sim_dir, output_dir=self.output_dir,
            problem_classes=[A, B, C]
        )
        t = TaskRunner(tasks=[task], scheduler=s)

        # Then
        self.assertEqual(len(t.todo), 5)
        # Basically only one instance of CommandTask should be created.
        names = [x.__class__.__name__ for x in t.todo]
        problems = [x.problem for x in t.todo if isinstance(x, SolveProblem)]
        self.assertEqual(names.count('RunAll'), 1)
        self.assertEqual(names.count('CommandTask'), 1)
        self.assertEqual(names.count('SolveProblem'), 3)
        self.assertEqual(len(problems), 3)
        self.assertEqual(
            sorted(x.__class__.__name__ for x in problems),
            ['A', 'B', 'C']
        )

        # When
        t.run(wait=0.1)

        # Then.
        self.assertEqual(t.todo, [])

        # The output dirs should exist now.
        for name in ('A', 'B', 'C'):
            self.assertTrue(
                os.path.exists(os.path.join(self.output_dir, name))
            )

        # When
        # We set force to True.
        task = RunAll(
            simulation_dir=self.sim_dir, output_dir=self.output_dir,
            problem_classes=[A, B, C], force=True
        )

        # Then
        # Make sure that all the output directories are deleted as this will
        for name in ('A', 'B', 'C'):
            self.assertFalse(
                os.path.exists(os.path.join(self.output_dir, name))
            )

    def test_problem_with_bad_requires_raises_error(self):
        # Given
        class D(Problem):
            def get_requires(self):
                return [('a', 'A')]

        # When
        self.assertRaises(
            RuntimeError,
            SolveProblem, D(self.sim_dir, self.output_dir)
        )

    def test_tasks_with_dependencies(self):
        # Given
        s = self._make_scheduler()
        cmd = 'python -c "import time; print(time.time())"'
        ct1_dir = os.path.join(self.sim_dir, '1')
        ct2_dir = os.path.join(self.sim_dir, '2')
        ct3_dir = os.path.join(self.sim_dir, '3')
        ct1 = CommandTask(cmd, output_dir=ct1_dir)
        ct2 = CommandTask(cmd, output_dir=ct2_dir, depends=[ct1])
        ct3 = CommandTask(cmd, output_dir=ct3_dir, depends=[ct1, ct2])

        # When
        t = TaskRunner(tasks=[ct1, ct2, ct3], scheduler=s)

        # Then
        self.assertEqual(len(t.todo), 3)

        # When
        t.run(wait=0.1)

        wait_until(lambda: not ct3.complete())

        # Then.
        # Ensure that the tasks are run in the right order.
        ct1_t, ct2_t, ct3_t = [
            self._get_time(x) for x in (ct1_dir, ct2_dir, ct3_dir)
        ]
        self.assertTrue(ct2_t > ct1_t)
        self.assertTrue(ct3_t > ct2_t)

    def test_simulation_with_dependencies(self):
        # Given
        class A(Problem):
            def setup(self):
                cmd = 'python -c "import time; print(time.time())"'
                s1 = Simulation(self.input_path('1'), cmd)
                s2 = Simulation(self.input_path('2'), cmd, depends=[s1])
                s3 = Simulation(self.input_path('3'), cmd, depends=[s1, s2])
                self.cases = [s1, s2, s3]

            def run(self):
                self.make_output_dir()

        s = self._make_scheduler()

        # When
        problem = A(self.sim_dir, self.output_dir)
        task = SolveProblem(problem)
        t = TaskRunner(tasks=[task], scheduler=s)

        # Then
        self.assertEqual(len(t.todo), 4)
        # Basically only one instance of CommandTask should be created.
        names = [x.__class__.__name__ for x in t.todo]
        self.assertEqual(names.count('CommandTask'), 3)
        self.assertEqual(names.count('SolveProblem'), 1)

        # When
        t.run(wait=0.1)
        wait_until(lambda: not task.complete())

        # Then
        ct1_t, ct2_t, ct3_t = [
            self._get_time(problem.input_path(x)) for x in ('1', '2', '3')
        ]
        self.assertTrue(ct2_t > ct1_t)
        self.assertTrue(ct3_t > ct2_t)


class TestLocalAutomation(TestAutomationBase):
    def _make_scheduler(self):
        worker = dict(host='localhost')
        s = Scheduler(root='.', worker_config=[worker])
        return s

    def test_automation(self):
        # Given.
        problem = EllipticalDrop(self.sim_dir, self.output_dir)
        s = self._make_scheduler()
        t = TaskRunner(tasks=[SolveProblem(problem=problem)], scheduler=s)

        # When.
        t.run(wait=1)

        # Then.
        sim1 = os.path.join(self.root, self.sim_dir,
                            'elliptical_drop', 'no_update_h')
        self.assertTrue(os.path.exists(sim1))
        sim2 = os.path.join(self.root, self.sim_dir,
                            'elliptical_drop', 'update_h')
        self.assertTrue(os.path.exists(sim2))

        results = os.path.join(self.root, self.output_dir,
                               'elliptical_drop', 'result.txt')
        self.assertTrue(os.path.exists(results))
        data = open(results).read()
        self.assertTrue('no_update_h' in data)
        self.assertTrue('update_h' in data)

        # When.
        problem = EllipticalDrop(self.sim_dir, self.output_dir)
        t = TaskRunner(tasks=[SolveProblem(problem=problem)], scheduler=s)

        # Then.
        self.assertEqual(len(t.todo), 0)

        # When
        problem.clean()

        # Then.
        out1 = os.path.join(self.root, self.output_dir,
                            'elliptical_drop', 'update_h')
        out2 = os.path.join(self.root, self.output_dir,
                            'elliptical_drop', 'no_update_h')
        self.assertFalse(os.path.exists(out1))
        self.assertFalse(os.path.exists(out2))
        self.assertTrue(os.path.exists(sim1))
        self.assertTrue(os.path.exists(sim2))

    def test_nothing_is_run_when_output_exists(self):
        # Given.
        s = self._make_scheduler()
        output = os.path.join(self.output_dir, 'elliptical_drop')
        os.makedirs(output)

        # When
        problem = EllipticalDrop(self.sim_dir, self.output_dir)
        t = TaskRunner(tasks=[SolveProblem(problem=problem)], scheduler=s)

        # Then.
        self.assertEqual(len(t.todo), 0)


class TestRemoteAutomation(TestLocalAutomation):
    def setUp(self):
        super(TestRemoteAutomation, self).setUp()
        self.other_dir = tempfile.mkdtemp()
        p = mock.patch.object(
            RemoteWorker, 'free_cores', return_value=2.0
        )
        p.start()
        self.addCleanup(p.stop)

    def tearDown(self):
        super(TestRemoteAutomation, self).tearDown()
        if os.path.exists(self.other_dir):
            safe_rmtree(self.other_dir)

    def _make_scheduler(self):
        workers = [
            dict(host='localhost'),
            dict(host='test_remote',
                 python=sys.executable, chdir=self.other_dir, testing=True)
        ]
        try:
            import execnet
        except ImportError:
            raise unittest.SkipTest('This test requires execnet')
        return Scheduler(root=self.sim_dir, worker_config=workers)

    def test_job_with_error_is_handled_correctly(self):
        # Given.
        problem = EllipticalDrop(self.sim_dir, self.output_dir)
        problem.cases[0].base_command += ' --xxx'
        s = self._make_scheduler()
        t = TaskRunner(tasks=[SolveProblem(problem=problem)], scheduler=s)

        # When.
        try:
            t.run(wait=1)
        except RuntimeError:
            pass

        # Then.

        # Ensure that the directories are copied over when they have errors.
        sim1 = os.path.join(self.root, self.sim_dir,
                            'elliptical_drop', 'no_update_h')
        self.assertTrue(os.path.exists(sim1))
        sim2 = os.path.join(self.root, self.sim_dir,
                            'elliptical_drop', 'update_h')
        self.assertTrue(os.path.exists(sim2))

        # Ensure that all the correct but already scheduled jobs are completed.
        task_status = t.task_status
        status_values = list(task_status.values())
        self.assertEqual(status_values.count('error'), 1)
        self.assertEqual(status_values.count('done'), 1)
        self.assertEqual(status_values.count('not started'), 1)
        for t, s in task_status.items():
            if s == 'done':
                self.assertTrue(t.complete())
            if s == 'error':
                self.assertRaises(RuntimeError, t.complete)


class TestCommandTask(TestAutomationBase):
    def _make_scheduler(self):
        worker = dict(host='localhost')
        s = Scheduler(root='.', worker_config=[worker])
        return s

    def test_command_tasks_executes_simple_command(self):
        # Given
        s = self._make_scheduler()
        cmd = 'python -c "print(1)"'
        t = CommandTask(cmd, output_dir=self.sim_dir)

        self.assertFalse(t.complete())

        # When
        t.run(s)
        wait_until(lambda: not t.complete())

        # Then
        self.assertTrue(t.complete())
        self.assertEqual(t.job_proxy.status(), 'done')
        self.assertEqual(t.job_proxy.get_stdout().strip(), '1')

    def test_command_tasks_converts_dollar_output_dir(self):
        # Given
        s = self._make_scheduler()
        cmd = '''python -c "print('$output_dir')"'''
        t = CommandTask(cmd, output_dir=self.sim_dir)

        self.assertFalse(t.complete())

        # When
        t.run(s)
        wait_until(lambda: not t.complete())

        # Then
        self.assertTrue(t.complete())
        self.assertEqual(t.job_proxy.status(), 'done')
        self.assertEqual(t.job_proxy.get_stdout().strip(), self.sim_dir)

    def test_command_tasks_handles_errors_correctly(self):
        # Given
        s = self._make_scheduler()
        cmd = 'python --junk'
        t = CommandTask(cmd, output_dir=self.sim_dir)

        self.assertFalse(t.complete())

        # When
        t.run(s)
        try:
            wait_until(lambda: not t.complete())
        except RuntimeError:
            pass

        # Then
        self.assertRaises(RuntimeError, t.complete)
        self.assertEqual(t.job_proxy.status(), 'error')

        # A new command task should still detect that the run failed, even
        # though the output directory exists.  In this case, it should
        # return False and not raise an error.

        # Given
        t = CommandTask(cmd, output_dir=self.sim_dir)

        # When/Then
        self.assertFalse(t.complete())


class TestFileCommandTask(TestAutomationBase):
    def _make_scheduler(self):
        worker = dict(host='localhost')
        s = Scheduler(root='.', worker_config=[worker])
        return s

    def test_file_command_tasks_works(self):
        # Given
        s = self._make_scheduler()
        pth = os.path.join(self.sim_dir, 'output.txt')
        cmd = 'python -c "from pathlib import Path; Path(%r).touch()"' % pth
        t = FileCommandTask(cmd, files=[pth])

        self.assertFalse(t.complete())

        # When
        t.run(s)
        wait_until(lambda: not t.complete())

        # Then
        self.assertTrue(t.complete())
        output_dir = pth + '.job_info'
        self.assertEqual(t.output_dir, output_dir)
        self.assertTrue(os.path.exists(t.output_dir))
        self.assertEqual(t.job_proxy.status(), 'done')
        self.assertTrue(os.path.exists(pth))


class TestRemoteCommandTask(TestAutomationBase):
    def _make_scheduler(self):
        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)
        worker = dict(
            host='test_remote', python=sys.executable,
            chdir=self.output_dir, testing=True
        )
        s = Scheduler(root='.', worker_config=[worker])
        return s

    def test_remote_command_tasks_complete_method_works(self):
        # Given
        s = self._make_scheduler()
        cmd = 'python -c "print(1)"'
        t = CommandTask(cmd, output_dir=self.sim_dir)

        self.assertFalse(t.complete())
        self.assertFalse(os.path.exists(
            os.path.join(self.sim_dir, 'stdout.txt')
        ))

        # When
        t.run(s)
        wait_until(lambda: not t.complete())

        # Then
        self.assertTrue(t.complete())
        self.assertTrue(os.path.exists(
            os.path.join(self.sim_dir, 'stdout.txt')
        ))
        # Test that if we call it repeatedly that it does indeed return True
        self.assertTrue(t.complete())


def test_simulation_get_labels():
    # Given
    s = Simulation(
        'junk', 'pysph run taylor_green',
        nx=25, perturb=0.1, correction=None
    )

    # When
    l = s.get_labels('nx')

    # Then
    assert l == r'nx=25'

    # When
    l = s.get_labels(['nx', 'perturb', 'correction'])

    # Then
    assert l == r'nx=25, perturb=0.1, correction'


class TestAutomator(TestAutomationBase):
    def setUp(self):
        super(TestAutomator, self).setUp()

    def test_automator_accepts_cluster_manager(self):
        # Given/When
        a = Automator('sim', 'output', [EllipticalDrop],
                      cluster_manager_factory=ClusterManager)

        # Then
        self.assertEqual(a.cluster_manager_factory, ClusterManager)

    @mock.patch.object(TaskRunner, 'run')
    def test_automator(self, mock_run):
        # Given
        a = Automator('sim', 'output', [EllipticalDrop])

        # When
        a.run([])

        # Then
        mock_run.assert_called_with()
        self.assertEqual(len(a.runner.todo), 4)

        expect = ['RunAll', 'SolveProblem', 'PySPHTask', 'PySPHTask']
        names = [x.__class__.__name__ for x in a.runner.todo]
        self.assertEqual(names, expect)

        # When
        # Given
        a = Automator('sim', 'output', [EllipticalDrop])
        a.run(['-m', '*no_up*'])

        # Then
        mock_run.assert_called_with()
        self.assertEqual(len(a.runner.todo), 3)

        expect = ['RunAll', 'SolveProblem', 'PySPHTask']
        names = [x.__class__.__name__ for x in a.runner.todo]
        self.assertEqual(names, expect)
        out_dir = os.path.basename(a.runner.todo[-1].output_dir)
        self.assertEqual(out_dir, 'no_update_h')

    @mock.patch.object(TaskRunner, 'run')
    def test_automates_only_tasks(self, mock_run):
        # Given
        a = Automator('sim', 'output', [])
        cmd = 'python -c "print(1)"'
        task = CommandTask(cmd, output_dir=self.sim_dir)
        a.add_task(task)

        # When
        a.run([])

        # Then
        mock_run.assert_called_with()
        self.assertEqual(len(a.runner.todo), 1)
        self.assertEqual(a.runner.todo[-1], task)

        # Given
        a = Automator('sim', 'output', [])
        cmd = 'python -c "print(1)"'
        task = CommandTask(cmd, output_dir=self.sim_dir)
        a.add_task(task, name='task')

        # When
        a.run([])

        # Then
        mock_run.assert_called_with()
        self.assertEqual(len(a.runner.todo), 0)

        # Given
        a = Automator('sim', 'output', [])
        cmd = 'python -c "print(1)"'
        task = CommandTask(cmd, output_dir=self.sim_dir)
        a.add_task(task, name='task')

        # When
        a.run(['task'])

        # Then
        mock_run.assert_called_with()
        self.assertEqual(len(a.runner.todo), 1)
        self.assertEqual(a.runner.todo[-1], task)

    @mock.patch.object(TaskRunner, 'run')
    def test_adding_problem_as_task(self, mock_run):
        # Given
        a = Automator('sim', 'output', [])
        a.add_task(EllipticalDrop)

        # When
        a.run([])

        # Then
        self.assertEqual(len(a.runner.todo), 3)

        expect = ['SolveProblem', 'PySPHTask', 'PySPHTask']
        names = [x.__class__.__name__ for x in a.runner.todo]
        self.assertEqual(names, expect)

        # Given
        a = Automator('sim', 'output', [])
        problem = EllipticalDrop('sim', 'output')
        a.add_task(problem)

        # When
        a.run([])

        # Then
        self.assertEqual(len(a.runner.todo), 3)

        expect = ['SolveProblem', 'PySPHTask', 'PySPHTask']
        names = [x.__class__.__name__ for x in a.runner.todo]
        self.assertEqual(names, expect)
        self.assertEqual(a.runner.todo[0].problem, problem)

    @mock.patch.object(TaskRunner, 'run')
    def test_automates_tasks_and_problems(self, mock_run):
        # Given
        a = Automator('sim', 'output', [EllipticalDrop])
        cmd = 'python -c "print(1)"'
        task = CommandTask(cmd, output_dir=self.sim_dir)
        a.add_task(task)
        dir2 = os.path.join(self.sim_dir, '2')
        task2 = CommandTask(cmd, output_dir=dir2)
        a.add_task(task2, post_proc=True)

        # When
        a.run([])

        # Then
        self.assertEqual(len(a.runner.todo), 6)

        expect = ['RunAll', 'SolveProblem', 'PySPHTask', 'PySPHTask',
                  'CommandTask', 'CommandTask']
        names = [x.__class__.__name__ for x in a.runner.todo]
        self.assertEqual(names, expect)
        self.assertEqual(a.runner.todo[-2], task)
        self.assertEqual(a.runner.todo[-1], task2)
        self.assertEqual(task2.depends, [a.runner.todo[0]])

    @mock.patch.object(TaskRunner, 'run')
    def test_automates_named_tasks(self, mock_run):
        # Given
        a = Automator('sim', 'output', [EllipticalDrop])
        cmd = 'python -c "print(1)"'
        task = CommandTask(cmd, output_dir=self.sim_dir)
        a.add_task(task, name='task')
        dir2 = os.path.join(self.sim_dir, '2')
        task2 = CommandTask(cmd, output_dir=dir2)
        a.add_task(task2, name='task2', post_proc=True)

        # When
        a.run([])

        # Then
        self.assertEqual(len(a.runner.todo), 4)

        expect = ['RunAll', 'SolveProblem', 'PySPHTask', 'PySPHTask']
        names = [x.__class__.__name__ for x in a.runner.todo]
        self.assertEqual(names, expect)
        self.assertEqual(task2.depends, [a.runner.todo[0]])

        # Given
        a = Automator('sim', 'output', [EllipticalDrop])
        cmd = 'python -c "print(1)"'
        task = CommandTask(cmd, output_dir=self.sim_dir)
        a.add_task(task, name='task')

        # When
        a.run(['task'])

        # Then
        self.assertEqual(len(a.runner.todo), 1)

        expect = ['CommandTask']
        names = [x.__class__.__name__ for x in a.runner.todo]
        self.assertEqual(names, expect)
        self.assertEqual(a.runner.todo[-1], task)

        # Given
        a = Automator('sim', 'output', [EllipticalDrop])
        cmd = 'python -c "print(1)"'
        task = CommandTask(cmd, output_dir=self.sim_dir)
        a.add_task(task, name='task')

        # When
        a.run(['all', 'task'])

        # Then
        self.assertEqual(len(a.runner.todo), 5)

        expect = ['RunAll', 'SolveProblem', 'PySPHTask', 'PySPHTask',
                  'CommandTask']
        names = [x.__class__.__name__ for x in a.runner.todo]
        self.assertEqual(names, expect)
        self.assertEqual(a.runner.todo[-1], task)



from __future__ import print_function

from fnmatch import fnmatch
import glob
import json
import os
import shlex
import shutil
import sys
import time
import traceback

from .jobs import Job


class Task(object):
    """Basic task to run.  Subclass this to do whatever is needed.

    This class is very similar to luigi's Task class.
    """
    def __init__(self, depends=None):
        # Depends is a list and available for all tasks.
        self.depends = depends if depends is not None else []

    def complete(self):
        """Should return True/False indicating success of task.

        If the task was just executed (in this invocation) but failed, raise
        any Exception that is a subclass of Exception as this signals an
        error to the task execution engine.

        If the task was executed in an earlier invocation of the automation,
        then just return True/False so as to be able to re-run the simulation.
        """
        return all([os.path.exists(x) for x in self.output()])

    def output(self):
        """Return list of output paths.
        """
        return []

    def run(self, scheduler):
        """Run the task, using the given scheduler.

        Using the scheduler is optional but recommended for any long-running
        tasks. It is safe to raise an exception immediately when running the
        task but for long running tasks, the exception will not matter and the
        `complete` method should do.
        """
        pass

    def requires(self):
        """Return iterable of tasks this task requires.

        It is important that one either return tasks that are idempotent or
        return the same instance as this method is called repeatedly.
        """
        return self.depends


class WrapperTask(Task):
    """A task that wraps other tasks and is done when all its requirements
    are done.
    """
    def complete(self):
        return all(r.complete() for r in self.requires())


class TaskRunner(object):
    """Run given tasks using the given scheduler.
    """
    def __init__(self, tasks, scheduler):
        """Constructor.

        **Parameters**

        tasks: iterable of `Task` instances.
        scheduler: `automan.jobs.Scheduler` instance
        """
        self.scheduler = scheduler
        self.todo = []
        self.task_status = dict()
        self.task_outputs = set()
        self.repeat_tasks = set()
        for task in tasks:
            self.add_task(task)

    # #### Private protocol  ##############################################

    def _check_error_in_running_tasks(self):
        running = self._get_tasks_with_status('running')
        for task in running:
            if self._check_status_of_task(task) == 'error':
                return True
        return False

    def _check_status_of_requires(self, task):
        status = [self._check_status_of_task(t) for t in task.requires()]

        if 'error' in status:
            return 'error'
        if all(x is True for x in status):
            return 'done'
        else:
            return 'running'

    def _check_status_of_task(self, task):
        status = self.task_status.get(task)
        if status == 'not started':
            return False
        elif status == 'done':
            return True
        else:
            complete = False
            try:
                complete = task.complete()
                self.task_status[task] = 'done' if complete else 'running'
            except Exception:
                complete = 'error'
                self.task_status[task] = 'error'
            return complete

    def _get_tasks_with_status(self, status):
        return [
            t for t, s in self.task_status.items()
            if s == status and t not in self.repeat_tasks
        ]

    def _is_output_registered(self, task):
        # Note, this has a side-effect of registering the task's output
        # when called.
        output = task.output()
        output_str = str(output)
        if output and output_str in self.task_outputs:
            self.repeat_tasks.add(task)
            return True
        else:
            if output:
                self.task_outputs.add(output_str)
            return False

    def _run(self, task):
        try:
            print("\nRunning task %s..." % task)
            self.task_status[task] = 'running'
            task.run(self.scheduler)
            status = 'running'
        except Exception:
            traceback.print_exc()
            status = 'error'
            self.task_status[task] = 'error'
        return status

    def _show_remaining_tasks(self, replace_line=False):
        start, end = ('\r', '') if replace_line else ('', '\n')
        running = self._get_tasks_with_status('running')
        print("{start}{pending} tasks pending and {running} tasks running".
              format(
                start=start, pending=len(self.todo), running=len(running)
              ), end=end)
        sys.stdout.flush()

    def _wait_for_running_tasks(self, wait):
        print("\nWaiting for already running tasks...")
        running = self._get_tasks_with_status('running')
        while len(running) > 0:
            for t in running:
                self._check_status_of_task(t)
            time.sleep(wait)
            running = self._get_tasks_with_status('running')
        errors = self._get_tasks_with_status('error')
        n_err = len(errors)
        print("{n_err} jobs had errors.".format(n_err=n_err))
        return n_err

    # #### Public protocol  ##############################################

    def add_task(self, task):
        if task in self.task_status or self._is_output_registered(task):
            # This task is already added or another task produces exactly
            # the same output, so do nothing.
            return

        if not task.complete():
            self.todo.append(task)
            self.task_status[task] = 'not started'
            for req in task.requires():
                self.add_task(req)
        else:
            self.task_status[task] = 'done'

    def run(self, wait=5):
        '''Run the tasks that were given.

        Wait for the given amount of time to poll for completed tasks.

        Returns the number of tasks that had errors.
        '''
        self._show_remaining_tasks()
        status = 'running'
        while len(self.todo) > 0 and status != 'error':
            to_remove = []
            for i in range(len(self.todo) - 1, -1, -1):
                task = self.todo[i]
                status = self._check_status_of_requires(task)
                if self._check_error_in_running_tasks():
                    status = 'error'

                if status == 'error':
                    break
                elif status == 'done':
                    to_remove.append(task)
                    status = self._run(task)

            for task in to_remove:
                self.todo.remove(task)

            if len(self.todo) > 0:
                self._show_remaining_tasks(replace_line=True)
                time.sleep(wait)

        n_errors = self._wait_for_running_tasks(wait)
        if n_errors == 0:
            print("Finished!")
        else:
            print("Please fix the issues and re-run.")
        return n_errors


class CommandTask(Task):
    """Convenience class to run a command via the framework. The class provides
    a method to run the simulation and also check if the simulation is
    completed. The command should ideally produce all of its outputs inside an
    output directory that is specified.

    """

    def __init__(self, command, output_dir, job_info=None, depends=None):
        """Constructor

        **Parameters**

        command: str or list: command to run; $output_dir is substituted.
        output_dir: str : path of output directory.
        job_info: dict: dictionary of job information.
        depends: list: list of tasks this depends on.

        """
        super().__init__(depends=depends)
        if isinstance(command, str):
            self.command = shlex.split(command)
        else:
            self.command = command
        self.command = [x.replace('$output_dir', output_dir)
                        for x in self.command]
        self.output_dir = output_dir
        self.job_info = job_info if job_info is not None else {}
        self.job_proxy = None
        self._copy_proc = None
        # This is a sentinel set to true when the job is finished
        # the data is copied to a local machine and cleaned on the remote.
        self._finished = False
        # This file will be created if the job exited with an error.
        self._error_status_file = os.path.join(
            self.output_dir, 'command_exited_with_error'
        )
        self._job = None

    def __str__(self):
        return ('%s, output in: %s ' %
                (self.__class__.__name__, self.output_dir))

    # #### Public protocol ###########################################

    def complete(self):
        """Should return True/False indicating success of task.
        """
        job_proxy = self.job_proxy
        if job_proxy is None:
            return self._is_done()
        elif self._finished:
            if os.path.exists(self._error_status_file):
                raise RuntimeError(
                    'Error in task with output in %s.' % self.output_dir
                )
            return True
        else:
            return self._copy_output_and_check_status()

    def run(self, scheduler):
        # Remove the error status file if it exists and we are going to run.
        if os.path.exists(self._error_status_file):
            os.remove(self._error_status_file)
        self.job_proxy = scheduler.submit(self.job)

    def clean(self):
        """Clean out any generated results.

        This completely removes the output directory.

        """
        if os.path.exists(self.output_dir):
            shutil.rmtree(self.output_dir)

    def output(self):
        """Return list of output paths.
        """
        return [self.output_dir]

    def requires(self):
        return self.depends

    # #### Private protocol ###########################################

    @property
    def job(self):
        if self._job is None:
            self._job = Job(
                command=self.command, output_dir=self.output_dir,
                **self.job_info
            )
        return self._job

    def _is_done(self):
        """Returns True if the simulation completed.
        """
        if (not os.path.exists(self.output_dir)) \
           or os.path.exists(self._error_status_file):
            return False
        else:
            return self.job.status() == 'done'

    def _check_if_copy_complete(self):
        proc = self._copy_proc
        if proc is None:
            # Local job so no copy needed.
            return True
        else:
            if proc.poll() is None:
                return False
            else:
                if self.job_proxy is not None:
                    self.job_proxy.clean()
                    self._finished = True
                return True

    def _copy_output_and_check_status(self):
        jp = self.job_proxy
        status = jp.status()
        if status == 'done':
            if self._copy_proc is None:
                self._copy_proc = jp.copy_output('.')
            return self._check_if_copy_complete()
        elif status == 'error':
            cmd = ' '.join(self.command)
            msg = '\n***************** ERROR *********************\n'
            msg += 'On host %s Job %s failed!' % (jp.worker.host, cmd)
            print(msg)
            print(jp.get_stderr())
            proc = jp.copy_output('.')
            if proc is not None:
                proc.wait()
            jp.clean()
            print('***************** ERROR **********************')
            with open(self._error_status_file, 'w') as fp:
                fp.write('')
            self._finished = True
            raise RuntimeError(msg)
        return False


class PySPHTask(CommandTask):
    """Convenience class to run a PySPH simulation via an automation
    framework.

    This task automatically adds the output directory specification for pysph
    so users to not need to add it.

    """

    def __init__(self, command, output_dir, job_info=None, depends=None):
        """Constructor

        **Parameters**

        command: str or list: command to run; $output_dir is substituted.
        output_dir: str : path of output directory.
        job_info: dict: dictionary of job information.
        depends: list: list of tasks this depends on.

        """
        super(PySPHTask, self).__init__(command, output_dir, job_info, depends)
        self.command += ['-d', output_dir]

    # #### Private protocol ###########################################

    def _is_done(self):
        """Returns True if the simulation completed.
        """
        if not os.path.exists(self.output_dir):
            return False
        job_status = self.job.status()
        if job_status == 'error':
            # If job information exists, it trumps everything else
            # as it stores the process exit status which is usually
            # a much better indicator of the job status.
            return False
        else:
            info_fname = self._get_info_filename()
            if not info_fname or not os.path.exists(info_fname):
                return False
            with open(info_fname) as fp:
                d = json.load(fp)
            return d.get('completed')

    def _get_info_filename(self):
        files = glob.glob(os.path.join(self.output_dir, '*.info'))
        if len(files) > 0:
            return files[0]
        else:
            return None


class FileCommandTask(CommandTask):
    """Convenience class to run a command which produces as output one or more
    files. The difference from the CommandTask is that this does not place its
    outputs in a separate directory.

    """
    def __init__(self, command, files, job_info=None, depends=None):
        """Constructor

        **Parameters**

        command: str or list: command to run; $output_dir is substituted.
        output_dir: str : path of output directory.
        files: list(str): relative paths of output files.
        job_info: dict: dictionary of job information.
        depends: list: list of tasks this depends on.

        """
        self.files = files
        output_dir = os.path.join(files[0] + '.job_info')
        super().__init__(
            command, output_dir, job_info=job_info, depends=depends
        )

    def clean(self):
        """Clean out any generated results.

        This completely removes the output directory.

        """
        if os.path.exists(self.output_dir):
            shutil.rmtree(self.output_dir)
        for f in self.files:
            if os.path.exists(f):
                os.remove(f)

    def output(self):
        """Return list of output paths.
        """
        return self.files


class Problem(object):
    """This class represents a numerical problem or computational
    problem of interest that needs to be solved.

    The class helps one run a variety of commands (or simulations),
    and then assemble/compare the results from those in the `run`
    method.  This is perhaps easily understood with an example.  Let
    us say one wishes to run the elliptical drop example problem with
    the standard SPH and TVF and compare the results and their
    convergence properties while also keep track of the computational
    time.  To do this one will have to run several simulations, then
    collect and process the results.  This is achieved by subclassing
    this class and implementing the following methods:

     - `get_name(self)`: returns a string of the name of the problem.  All
       results and simulations are collected inside a directory with
       this name.
     - `get_commands(self)`: returns a sequence of (directory_name,
       command_string, job_info, depends) tuples. These are to be executed
       before the `run` method is called.
     - `get_requires(self)`: returns a sequence of (name, task) tuples. These
       are to be exeuted before the `run` method is called.
     - `run(self)`: Processes the completed simulations to make plots etc.

    See the `EllipticalDrop` example class below to see a full implementation.

    """

    # The Task class to create for the cases, change to suit your needs.
    task_cls = CommandTask

    def __init__(self, simulation_dir, output_dir):
        """Constructor.

        **Parameters**

        simulation_dir : str : directory where simulation output goes.
        output_dir : str : directory where outputs from `run` go.
        """
        self.out_dir = output_dir
        self.sim_dir = simulation_dir

        # Setup the simulation instances in the cases.
        self.cases = None
        self.setup()

    def _make_depends(self, depends):
        if not depends:
            return []
        deps = []
        for x in depends:
            if isinstance(x, Task):
                deps.append(x)
            elif isinstance(x, Simulation):
                if x.depends:
                    my_depends = self._make_depends(x.depends)
                else:
                    my_depends = None
                task = self.task_cls(
                    x.command, self.input_path(x.name), x.job_info,
                    depends=my_depends
                )
                deps.append(task)
            else:
                raise RuntimeError(
                    'Invalid dependency: {0} for problem {1}'.format(
                        x, self
                    )
                )

        return deps

    # #### Public protocol ###########################################

    def input_path(self, *args):
        """Given any arguments, relative to the simulation dir, return
        the absolute path.
        """
        return os.path.join(self.sim_dir, self.get_name(), *args)

    simulation_path = input_path

    def output_path(self, *args):
        """Given any arguments relative to the output_dir return the
        absolute path.
        """
        return os.path.join(self.out_dir, self.get_name(), *args)

    def setup(self):
        """Called by init, so add any initialization here.
        """
        pass

    def make_output_dir(self):
        """Convenience to make the output directory if needed.
        """
        base = self.output_path()
        if not os.path.exists(base):
            os.makedirs(base)

    def get_name(self):
        """Return the name of this problem, this name is used as a
        directory for the simulation and the outputs.
        """
        # Return a sane default instead of forcing the user to do this.
        return self.__class__.__name__

    def get_commands(self):
        """Return a sequence of (name, command_string, job_info_dict)
        or (name, command_string, job_info_dict, depends).

        The name represents the command being run and is used as a subdirectory
        for generated output.

        The command_string is the command that needs to be run.

        The job_info_dict is a dictionary with any additional info to be used
        by the job, these are additional arguments to the
        `automan.jobs.Job` class. It may be None if nothing special need
        be passed.

        The depends is any dependencies this simulation has in terms of other
        simulations/tasks.

        """
        if self.cases is not None:
            return [
                (x.name, x.command, x.job_info, x.depends) for x in self.cases
            ]
        else:
            return []

    def get_requires(self):
        """Return a sequence of tuples of form (name, task).

        The name represents the command being run and is used as
        a subdirectory for generated output.

        The task is a `automan.automation.Task` instance.
        """
        base = self.get_name()
        result = []
        for cmd_info in self.get_commands():
            name, cmd, job_info = cmd_info[:3]
            deps = cmd_info[3] if len(cmd_info) == 4 else []
            sim_output_dir = self.input_path(name)
            depends = self._make_depends(deps)
            task = self.task_cls(
                cmd, sim_output_dir, job_info, depends=depends
            )
            task_name = '%s.%s' % (base, name)
            result.append((task_name, task))
        return result

    def get_outputs(self):
        """Get a list of outputs generated by this problem.  By default it
        returns the output directory (as a single element of a list).
        """
        return [self.output_path()]

    def run(self):
        """Run any analysis code for the simulations completed.  This
        is usually run after the simulation commands are completed.
        """
        pass

    def clean(self):
        """Cleanup any generated output from the analysis code.  This does not
        clean the output of any nested commands.
        """
        for path in self.get_outputs():
            if os.path.exists(path):
                if os.path.isdir(path):
                    shutil.rmtree(path)
                elif os.path.isfile(path):
                    os.remove(path)


class PySPHProblem(Problem):
    task_cls = PySPHTask


def key_to_option(key):
    """Convert a dictionary key to a valid command line option.  This simply
    replaces underscores with dashes.
    """
    return key.replace('_', '-')


def kwargs_to_command_line(kwargs):
    """Convert a dictionary of keyword arguments to a list of command-line
    options.  If the value of the key is None, no value is passed.

    **Examples**

    >>> sorted(kwargs_to_command_line(dict(some_arg=1, something_else=None)))
    ['--some-arg=1', '--something-else']
    """
    cmd_line = []
    for key, value in kwargs.items():
        option = key_to_option(key)
        if value is None:
            arg = "--{option}".format(option=option)
        else:
            arg = "--{option}={value}".format(
                option=option, value=str(value)
            )

        cmd_line.append(arg)
    return cmd_line


class Simulation(object):
    """A convenient class to abstract code for a particular simulation.
    Simulation objects are typically created by ``Problem`` instances in order
    to abstract and simulate repetitive code for a particular simulation.

    For example if one were comparing the elliptical_drop example, one could
    instantiate a Simulation object as follows::

        >>> s = Simlation('outputs/sph', 'pysph run elliptical_drop')

    One can pass any additional command line arguments as follows::

        >>> s = Simlation(
        ...     'outputs/sph', 'pysph run elliptical_drop', timestep=0.005
        ... )
        >>> s.command
        'pysph run elliptical_drop --timestep=0.001'
        >>> s.input_path('results.npz')
        'outputs/sph/results.npz'

    The extra parameters can be used to filter and compare different
    simulations.  One can define additional plot methods for a particular
    subclass and use these to easily plot results for different cases.

    One can also pass any additional parameters to the `automan.jobs.Job`
    class via the job_info kwarg so as to run the command suitably. For
    example::

        >>> s = Simlation('outputs/sph', 'pysph run elliptical_drop',
        ...               job_info=dict(n_thread=4))

    The object has other methods that are convenient when comparing plots.
    Along with the ``compare_cases``, ``filter_cases`` and ``filter_by_name``
    this is an extremely powerful way to automate and compare results.

    """
    def __init__(self, root, base_command, job_info=None, depends=None, **kw):
        """Constructor

        **Parameters**

        root: str
            Path to simulation output directory.
        base_command: str
            Base command to run.
        job_info: dict
            Extra arguments to the `automan.jobs.Job` class.
        depends: list
            List of other simulations/tasks this simulation depends on.
        **kw: dict
            Additional parameters to pass to command.
        """
        self.root = root
        self.name = os.path.basename(root)
        self.base_command = base_command
        self.job_info = job_info
        self.depends = depends if depends is not None else []
        self.params = dict(kw)
        self._results = None

    def input_path(self, *args):
        """Given any arguments, relative to the simulation dir, return
        the absolute path.
        """
        return os.path.join(self.root, *args)

    @property
    def command(self):
        return self.base_command + ' ' + self.get_command_line_args()

    @property
    def data(self):
        if self._results is None:
            import numpy
            self._results = numpy.load(self.input_path('results.npz'))
        return self._results

    def get_labels(self, labels):
        render = self.render_parameter
        if isinstance(labels, str):
            return render(labels)
        else:
            s = [render(x) for x in labels]
            s = [x for x in s if len(x) > 0]
            return r', '.join(s)

    def kwargs_to_command_line(self, kwargs):
        return kwargs_to_command_line(kwargs)

    def get_command_line_args(self):
        return ' '.join(self.kwargs_to_command_line(self.params))

    def render_parameter(self, param):
        """Return string to be used for labels for given parameter.
        """
        if param not in self.params:
            return ''
        value = self.params[param]
        if value is None:
            return r'%s' % param
        else:
            return r'%s=%s' % (param, self.params[param])


############################################################################
# Convenient classes that can be used to easily automate a collection
# of problems.

class SolveProblem(Task):
    """Solves a particular `Problem`. This runs all the commands that the
    problem requires and then runs the problem instance's run method.

    The match argument is a string which when provided helps run only a subset
    of the requirements for the problem.

    The force argument specifies that the problem should be cleaned, so as to
    re-run any post-processing.
    """

    def __init__(self, problem, match='', force=False, depends=None):
        super().__init__(depends=depends)
        self.problem = problem
        self.match = match
        self.force = force
        if self.force:
            self.problem.clean()
        self._requires = [
            self._make_task(task)
            for name, task in self.problem.get_requires()
            if len(match) == 0 or fnmatch(name, match)
        ]

    def _make_task(self, obj):
        if isinstance(obj, Task):
            return obj
        elif isinstance(obj, Problem):
            return SolveProblem(
                problem=obj, match=self.match, force=self.force
            )
        elif isinstance(obj, type) and issubclass(obj, Problem):
            problem = obj(self.problem.sim_dir, self.problem.out_dir)
            return SolveProblem(
                problem=problem, match=self.match, force=self.force
            )
        else:
            raise RuntimeError(
                'Unknown requirement: {0}, for problem: {1}.'.format(
                    obj, self.problem
                )
            )

    def __str__(self):
        return 'Problem named %s' % self.problem.get_name()

    def complete(self):
        if len(self.match) == 0:
            return super(SolveProblem, self).complete()
        else:
            return all(r.complete() for r in self.requires())

    def output(self):
        return self.problem.get_outputs()

    def run(self, scheduler):
        if len(self.match) == 0:
            self.problem.run()

    def requires(self):
        return self._requires + self.depends


class RunAll(WrapperTask):
    """Solves a given collection of problems.
    """

    def __init__(self, simulation_dir, output_dir, problem_classes,
                 force=False, match='', depends=None):
        super().__init__(depends=depends)
        self.simulation_dir = simulation_dir
        self.output_dir = output_dir
        self.force = force
        self.match = match
        self.problems = self._make_problems(problem_classes)
        self._requires = self._get_requires()

    # #### Private protocol  ###############################################

    def _get_requires(self):
        return [
            SolveProblem(problem=x, match=self.match, force=self.force)
            for x in self.problems
        ]

    def _make_problems(self, problem_classes):
        problems = []
        for klass in problem_classes:
            problem = klass(self.simulation_dir, self.output_dir)
            problems.append(problem)
        return problems

    # #### Public protocol  ################################################

    def requires(self):
        return self._requires + self.depends


class Automator(object):
    """Main class to automate a collection of problems.

    This processess command line options and runs all tasks with a scheduler
    that is configured using the ``config.json`` file if it is present. Here is
    typical usage::

        >>> all_problems = [EllipticalDrop]
        >>> automator = Automator('outputs', 'figures', all_problems)
        >>> automator.run()

    The class also creates a `automan.cluster_manager.ClusterManager`
    instance and integrates the cluster management features as well. This
    allows a user to automate their results across a collection of remote
    machines accessible only by ssh.

    """
    def __init__(self, simulation_dir, output_dir, all_problems,
                 cluster_manager_factory=None):
        """Constructor.

        **Parameters**

        simulation_dir : str
            Root directory to generate simulation results in.
        output_dir: str
            Root directory where outputs will be generated by Problem
            instances.
        all_problems: sequence of `Problem` classes.
            Sequence of problem classes to automate.
        cluster_manager_factory: callable
            Callable should return `cluster_manager.ClusterManager` instance.
            None will use the default one.
        """
        self.simulation_dir = simulation_dir
        self.output_dir = output_dir
        self.all_problems = all_problems
        self.named_tasks = {}
        self.tasks = []
        self.post_proc_tasks = []
        self.runner = None
        self.cluster_manager = None
        self.runall_task = None
        self._args = None
        if cluster_manager_factory is None:
            from automan.cluster_manager import ClusterManager
            self.cluster_manager_factory = ClusterManager
        else:
            self.cluster_manager_factory = cluster_manager_factory

    # #### Public Protocol ########################################

    def add_task(self, task, name=None, post_proc=False):
        """Add a task or a problem instance to also execute.

        If the `name` is specified then it is a treated as a named task wherein
        it must be only invoked explicitly via the command line when asked.

        If `post_proc` is True then the task is given an additional dependency
        if possible such that the task is run after the `RunAll` task is
        completed.

        **Parameters**

        task: Task or Problem instance: Task or Problem to add.
        name: str: name of the task (optional).
        post_proc: bool: Add a dependency to the task with the RunAll task.

        """
        if isinstance(task, type) and issubclass(task, Problem):
            p = task(
                simulation_dir=self.simulation_dir, output_dir=self.output_dir
            )
            _task = SolveProblem(p)
        elif isinstance(task, Problem):
            _task = SolveProblem(task)
        elif isinstance(task, Task):
            _task = task
        else:
            raise ValueError(
                'Invalid task: must be Problem class/instance or Task.'
            )
        if name is not None:
            self.named_tasks[name] = _task
        else:
            self.tasks.append(_task)

        if post_proc:
            self.post_proc_tasks.append(_task)

    def run(self, argv=None):
        """Start the automation.
        """
        self._setup(argv)
        self._setup_tasks()
        self.runner.run()

    # #### Private Protocol ########################################

    def _check_positional_arguments(self, problems):
        names = [c.__name__ for c in self.all_problems] + ['all']
        lower_names = [x.lower() for x in names]
        lower_names.extend(list(self.named_tasks.keys()))
        for p in problems:
            if p.lower() not in lower_names:
                print("ERROR: %s not a valid problem/task!" % p)
                print("Valid names are %s" % ', '.join(names))
                self.parser.exit(1)

    def _get_exclude_paths(self):
        """Returns a list of exclude paths suitable for passing on to rsync to
        exclude syncing some directories on remote machines.
        """
        paths = []
        for path in [self.simulation_dir, self.output_dir]:
            if not path.endswith('/'):
                paths.append(path + '/')
        return paths

    def _parse_args(self, argv):
        '''Parse command line arguments.

        Override this when the CLI arguments are customized.
        '''
        self._args = self.parser.parse_args(argv)
        return self._args

    def _select_problem_classes(self, problems):
        if 'all' in problems:
            return self.all_problems
        else:
            lower_names = [x.lower() for x in problems]
            return [cls for cls in self.all_problems
                    if cls.__name__.lower() in lower_names]

    def _setup(self, argv):
        if self.runner is None:
            self._setup_argparse()
            args = self._parse_args(argv)

            self._check_positional_arguments(args.problem)

            self.cluster_manager = self.cluster_manager_factory(
                config_fname=args.config,
                exclude_paths=self._get_exclude_paths()
            )
            from .cluster_manager import BootstrapError

            if len(args.host) > 0:
                try:
                    self.cluster_manager.add_worker(
                        args.host, args.home, args.nfs
                    )
                except BootstrapError:
                    pass
                return
            elif len(args.host) == 0 and args.update_remote:
                self.cluster_manager.update(not args.no_rebuild)
            elif len(args.rm_remote_output) > 0:
                self.cluster_manager.delete(
                    self.simulation_dir, args.rm_remote_output)

            problem_classes = self._select_problem_classes(args.problem)
            task = RunAll(
                simulation_dir=self.simulation_dir,
                output_dir=self.output_dir,
                problem_classes=problem_classes,
                force=args.force, match=args.match
            )
            self.runall_task = task

            self.scheduler = self.cluster_manager.create_scheduler()
            self.runner = TaskRunner([task], self.scheduler)

    def _setup_argparse(self):
        import argparse
        desc = "Automation script to run simulations."
        parser = argparse.ArgumentParser(
            description=desc
        )
        all_problem_names = [c.__name__ for c in self.all_problems]
        all_problem_names += list(self.named_tasks.keys()) + ['all']
        parser.add_argument(
            'problem', nargs='*', default=["all"],
            help="Specifies problem/task to run as a string "
            "(case-insensitive), valid names are %s.  "
            "Defaults to running all of the problems."
            % all_problem_names
        )

        parser.add_argument(
            '-a', '--add-node', action="store", dest="host", type=str,
            default='', help="Add a new remote worker."
        )
        parser.add_argument(
            '-c', '--config', action="store", dest="config",
            default="config.json", help="Configuration file to use."
        )
        parser.add_argument(
            '--home', action="store", dest="home", type=str,
            default='',
            help='Home directory of the remote worker (to be used with -a)'
        )
        parser.add_argument(
            '--nfs', action="store_true", dest="nfs",
            default=False,
            help=('Does the remote remote worker share the filesystem '
                  '(to be used with -a)')
        )
        parser.add_argument(
            '-f', '--force', action="store_true", default=False, dest='force',
            help='Redo the plots even if they were already made.'
        )
        parser.add_argument(
            '-m', '--match', action="store", type=str, default='',
            dest='match', help="Name of the problem to run (uses fnmatch)"
        )
        parser.add_argument(
            '--no-rebuild', action="store_true",
            dest="no_rebuild", default=False,
            help="Do not rebuild the sources on update, just update the files."
        )
        parser.add_argument(
            '-u', '--update-remote', action='store_true',
            dest='update_remote', default=False,
            help='Update remote worker machines.'
        )
        parser.add_argument(
            '--rm-remote-output', nargs='*', action="store",
            dest="rm_remote_output", type=str, default='',
            help="remove output folder from the mentioned machines use"
            "'all' to remove from all host (except localhost and host where "
            "nfs is true)"
        )

        self.parser = parser

    def _setup_tasks(self):
        for task in self.post_proc_tasks:
            task.depends.append(self.runall_task)

        # Add generic tasks.
        for task in self.tasks:
            self.runner.add_task(task)

        # Add named tasks only if specifically requested on CLI.
        for name, task in self.named_tasks.items():
            if name in self._args.problem:
                self.runner.add_task(task)

        # Reset the tasks so we can use the automator interactively.
        self.post_proc_tasks = []
        self.tasks = []
        self.named_tasks = {}



from unittest import mock
from automan.automation import Simulation
from automan.utils import (compare_runs, dprod, filter_cases, mdict,
                           opts2path)


def test_compare_runs_calls_methods_when_given_names():
    # Given
    sims = [mock.MagicMock(), mock.MagicMock()]
    s0, s1 = sims
    s0.get_labels.return_value = s1.get_labels.return_value = 'label'

    # When
    compare_runs(sims, 'fig', labels=['x'], exact='exact')

    # Then
    s0.exact.assert_called_once_with(color='k', linestyle='-')
    s0.fig.assert_called_once_with(color='k', label='label', linestyle='--')
    s0.get_labels.assert_called_once_with(['x'])
    assert s1.exact.called is False
    s1.fig.assert_called_once_with(color='k', label='label', linestyle='-.')
    s1.get_labels.assert_called_once_with(['x'])


def test_compare_runs_works_when_given_callables():
    # Given
    sims = [mock.MagicMock()]
    s0 = sims[0]
    s0.get_labels.return_value = 'label'

    func = mock.MagicMock()
    exact = mock.MagicMock()

    # When
    compare_runs(sims, func, labels=['x'], exact=exact)

    # Then
    exact.assert_called_once_with(s0, color='k', linestyle='-')
    func.assert_called_once_with(s0, color='k', label='label', linestyle='--')
    s0.get_labels.assert_called_once_with(['x'])


def test_compare_runs_uses_given_styles():
    # Given
    sims = [mock.MagicMock()]
    s0 = sims[0]
    s0.get_labels.return_value = 'label'

    func = mock.MagicMock()
    styles = mock.MagicMock()
    styles.return_value = iter([dict(linestyle=':', color='b')])

    # When
    compare_runs(sims, func, labels=['x'], styles=styles)

    # Then
    func.assert_called_once_with(s0, color='b', label='label', linestyle=':')
    s0.get_labels.assert_called_once_with(['x'])
    styles.assert_called_once_with(sims)


def test_compare_runs_uses_given_styles_returning_iterable():
    # Given
    sims = [mock.MagicMock()]
    s0 = sims[0]
    s0.get_labels.return_value = 'label'

    func = mock.MagicMock()
    styles = mock.MagicMock()
    # Return an iterable and not an iterator
    styles.return_value = [dict(linestyle=':', color='b')]

    # When
    compare_runs(sims, func, labels=['x'], styles=styles)

    # Then
    func.assert_called_once_with(s0, color='b', label='label', linestyle=':')
    s0.get_labels.assert_called_once_with(['x'])
    styles.assert_called_once_with(sims)


def test_dprod():
    # Given/When
    res = dprod(mdict(a=[1, 2], b=['xy']), mdict(c='ab'))
    # Then
    exp = [{'a': 1, 'b': 'xy', 'c': 'a'},
           {'a': 1, 'b': 'xy', 'c': 'b'},
           {'a': 2, 'b': 'xy', 'c': 'a'},
           {'a': 2, 'b': 'xy', 'c': 'b'}]
    assert res == exp


def test_filter_cases_works_with_params():
    # Given
    sims = [Simulation(root='', base_command='python', param1=i, param2=i+1)
            for i in range(5)]
    # When
    result = filter_cases(sims, param1=2)

    # Then
    assert len(result) == 1
    assert result[0].params['param1'] == 2

    # When
    result = filter_cases(sims, param1=2, param2=2)

    # Then
    assert len(result) == 0

    # When
    result = filter_cases(sims, param1=3, param2=4)

    # Then
    assert len(result) == 1
    assert result[0].params['param1'] == 3
    assert result[0].params['param2'] == 4


def test_filter_cases_works_with_predicate():
    # Given
    sims = [Simulation(root='', base_command='python', param1=i, param2=i+1)
            for i in range(5)]

    # When
    result = filter_cases(
        sims, predicate=lambda x: x.params.get('param1', 0) % 2
    )

    # Then
    assert len(result) == 2
    assert result[0].params['param1'] == 1
    assert result[1].params['param1'] == 3

    # When
    result = filter_cases(
        sims, predicate=2
    )

    # Then
    assert len(result) == 0

    # Given
    sims = [Simulation(root='', base_command='python', predicate=i)
            for i in range(5)]

    # When
    result = filter_cases(
        sims, predicate=2
    )

    # Then
    assert len(result) == 1
    assert result[0].params['predicate'] == 2


def test_mdict():
    exp = [{'a': 1, 'b': 'x'},
           {'a': 1, 'b': 'y'},
           {'a': 2, 'b': 'x'},
           {'a': 2, 'b': 'y'}]
    assert mdict(a=[1, 2], b='xy') == exp


def test_opts2path():
    assert opts2path(dict(x=1, y='hello', z=0.1)) == 'x_1_hello_z_0.1'
    assert opts2path(dict(x=1, y='hello', z=0.1), keys=['x']) == 'x_1'
    res = opts2path(dict(x=1, y='hello', z=0.1), ignore=['x'])
    assert res == 'hello_z_0.1'
    res = opts2path(dict(x=1, y='hello', z=0.1), kmap=dict(x='XX'))
    assert res == 'XX_1_hello_z_0.1'



import json
import os
import sys


def write_pysph_data(directory):
    info_fname = os.path.join(directory, 'example.info')
    data = {
        'completed': True,
        'output_dir': directory,
        'fname': 'example'
    }
    with open(info_fname, 'w') as fp:
        json.dump(data, fp)

    with open(os.path.join(directory, 'results.dat'), 'w') as fp:
        fp.write(str(sys.argv[1:]))


def main():
    from argparse import ArgumentParser
    p = ArgumentParser()
    p.add_argument(
        '-d', '--directory', default='example_output',
        help='Output directory'
    )
    p.add_argument(
        '--update-h', action="store_true", dest='update_h',
        help='Update h'
    )
    p.add_argument(
        '--no-update-h', action="store_false", dest='update_h',
        help='Do not update h'
    )
    o = p.parse_args()

    if not os.path.exists(o.directory):
        os.mkdir(o.directory)

    write_pysph_data(o.directory)


if __name__ == '__main__':
    main()



from automan.api import Problem, Automator


class Squares(Problem):
    def get_name(self):
        return 'squares'

    def get_commands(self):
        return [
            ('1', 'python square.py 1', None),
            ('2', 'python square.py 2', None),
        ]

    def run(self):
        self.make_output_dir()


if __name__ == '__main__':
    automator = Automator(
        simulation_dir='outputs',
        output_dir='manuscript/figures',
        all_problems=[Squares]
    )
    automator.run()



from .jobs import Job, Worker, LocalWorker, RemoteWorker, Scheduler  # noqa

from .automation import (  # noqa
    Automator, CommandTask, FileCommandTask, Problem, PySPHProblem, PySPHTask,
    RunAll, Simulation, SolveProblem, Task, TaskRunner, WrapperTask
)

from .utils import ( # noqa
    compare_runs, dprod, filter_by_name, filter_cases, mdict,
    opts2path
)

from .cluster_manager import ClusterManager  # noqa
from .conda_cluster_manager import CondaClusterManager  # noqa
from .edm_cluster_manager import EDMClusterManager  # noqa



from automan.api import Problem, Automator
from matplotlib import pyplot as plt
import numpy as np


class Squares(Problem):
    def get_name(self):
        return 'squares'

    def get_commands(self):
        commands = [(str(i), 'python square.py %d' % i, None)
                    for i in range(1, 8)]
        return commands

    def run(self):
        self.make_output_dir()
        data = []
        for i in range(1, 8):
            stdout = self.input_path(str(i), 'stdout.txt')
            with open(stdout) as f:
                values = [float(x) for x in f.read().split()]
                data.append(values)

        data = np.asarray(data)
        plt.plot(data[:, 0], data[:, 1], 'o-')
        plt.xlabel('x')
        plt.ylabel('y')
        plt.savefig(self.output_path('squares.pdf'))
        plt.close()


from automan.api import Simulation


class Powers(Problem):
    def get_name(self):
        return 'powers'

    def setup(self):
        base_cmd = 'python powers.py --output-dir $output_dir'
        self.cases = [
            Simulation(
                root=self.input_path(str(i)),
                base_command=base_cmd,
                power=float(i)
            )
            for i in range(1, 5)
        ]

    def run(self):
        self.make_output_dir()
        plt.figure()
        for case in self.cases:
            data = np.load(case.input_path('results.npz'))
            plt.plot(
                data['x'], data['y'],
                label=r'$x^{{%.2f}}$' % case.params['power']
            )
        plt.grid()
        plt.xlabel('x')
        plt.ylabel('y')
        plt.legend()
        plt.savefig(self.output_path('powers.pdf'))
        plt.close()


if __name__ == '__main__':
    automator = Automator(
        simulation_dir='outputs',
        output_dir='manuscript/figures',
        all_problems=[Squares, Powers]
    )
    automator.run()



from setuptools import setup, find_packages
import sys


def get_version():
    import os
    data = {}
    fname = os.path.join('automan', '__init__.py')
    exec(compile(open(fname).read(), fname, 'exec'), data)
    return data.get('__version__')


install_requires = ['psutil', 'execnet', 'setuptools']
tests_require = ['pytest']
if sys.version_info.major < 3:
    tests_require.append('mock')
    install_requires.append('mock')

classes = """
Development Status :: 3 - Alpha
Environment :: Console
Intended Audience :: Developers
Intended Audience :: Education
Intended Audience :: End Users/Desktop
Intended Audience :: Science/Research
License :: OSI Approved :: BSD License
Natural Language :: English
Operating System :: OS Independent
Programming Language :: Python
Programming Language :: Python :: 2.7
Programming Language :: Python :: 3
Programming Language :: Python :: 3.3
Programming Language :: Python :: 3.4
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
Topic :: Scientific/Engineering
Topic :: Software Development :: Libraries
Topic :: Utilities
"""

classifiers = [x.strip() for x in classes.splitlines() if x]

setup(
    name='automan',
    version=get_version(),
    author='Prabhu Ramachandran',
    author_email='prabhu@aero.iitb.ac.in',
    description='A simple Python-based automation framework.',
    long_description=open('README.rst').read(),
    license="BSD",
    url='https://github.com/pypr/automan',
    classifiers=classifiers,
    packages=find_packages(),
    install_requires=install_requires,
    tests_require=tests_require,
    package_dir={'automan': 'automan'},
)



"""Code to bootstrap and update the project so a remote host can be used as a
worker to help with the automation of tasks.

This requires ssh/scp and rsync to work on all machines.

This is currently only tested on Linux machines.

"""

import json
import os
import shlex
import stat
import subprocess
import sys
from textwrap import dedent

try:
    from urllib import urlopen
except ImportError:
    from urllib.request import urlopen


class BootstrapError(Exception):
    pass


class ClusterManager(object):
    """The cluster manager class.

    This class primarily helps setup software on a remote worker machine such
    that it can run any computational jobs from the automation framework.

    The general directory structure of a remote worker machine is as follows::

        remote_home/           # Could be ~
            automan/           # Root of automation directory (configurable)
                envs/          # python virtual environments for use.
                my_project/    # Current directory for specific projects.

    The project directories are synced from this machine to the remote worker.

    The "my_project" is the root of the directory with the automation script
    and this should contain the required sources that need to be executed. One
    can use a list of source directories which will be copied over but it is
    probably most convenient to put it all in the root of the project directory
    to keep everything self-contained.

    The `ClusterManager` class manages these remote workers by helping setup
    the directories, bootstrapping the Python virtualenv and also keeping these
    up-to-date as project directory is changed on the local machine.

    The class therefore has two primary public methods,

    1. `add_worker(self, host, home, nfs)` which adds a new worker machine by
       bootstrapping the machine with the software and the appropriate source
       directories.

    2. `update()`, which keeps the directory and software up-to-date.

    The class variables BOOTSTRAP and UPDATE are the content of scripts
    uploaded to these machines and should be extended by users to do what they
    wish.

    The class creates a ``config.json`` in the current working directory that
    may be edited by a user. It also creates a directory called
    ``.{self.root}`` which defaults to ``.automan``. The bootstrap and
    update scripts are put here and may be edited by the user for any new
    hosts.

    One may override the `_get_python, _get_helper_scripts`, and
    `_get_bootstrap_code, _get_update_code` methods to change this to use other
    package managers like edm or conda. See the conda_cluster_manager for an
    example.

    """

    #######################################################
    # These scripts are used to bootstrap the installation
    # and update them.
    BOOTSTRAP = dedent("""\
        #!/bin/bash

        set -e
        if hash virtualenv 2>/dev/null; then
            virtualenv -p python3 --system-site-packages envs/{project_name}
        else
            python3 virtualenv.pyz --system-site-packages envs/{project_name}
        fi
        source envs/{project_name}/bin/activate

        pip install automan

        # Run any requirements.txt from the user
        cd {project_name}
        if [ -f "requirements.txt" ] ; then
            pip install -r requirements.txt
        fi
        """)

    UPDATE = dedent("""\
         #!/bin/bash

         set -e
         source envs/{project_name}/bin/activate
         # Run any requirements.txt from the user
         cd {project_name}
         if [ -f "requirements.txt" ] ; then
             pip install -r requirements.txt
         fi
         """)
    #######################################################

    def __init__(self, root='automan', sources=None,
                 config_fname='config.json', exclude_paths=None,
                 testing=False):
        """Create a cluster manager instance.

        **Parameters**

        root: str
           The name of the root directory where all the files on the remote
           will be created.
        sources: list
           A list of source directories to sync.
        config_fname: str
           The name of the config file to create.
        exclude_paths: list
           A list of paths to exclude while syncing. This is in a form suitable
           to pass to rsync.
        testing: bool
           Use this while testing. This allows us to run unit tests for remotes
           on the local machine.

        """
        self.root = root
        self.workers = []
        self.sources = sources
        self.scripts_dir = os.path.abspath('.' + self.root)
        self.exclude_paths = exclude_paths if exclude_paths else []
        self.testing = testing

        # This is setup by the config and is the name of
        # the project directory.
        self.project_name = None

        # The config file will always trump any direct settings
        # unless there is no config file.
        self.config_fname = config_fname
        self._read_config()
        if not os.path.exists(self.scripts_dir):
            os.makedirs(self.scripts_dir)

    # ### Private Protocol ########################################
    def _bootstrap(self, host, home):
        helper_scripts = self._get_helper_scripts()
        base_cmd = ("cd {home}; mkdir -p {root}/envs; "
                    "mkdir -p {root}/{project_name}/.{root}").format(
                        home=home, root=self.root,
                        project_name=self.project_name
                    )
        self._ssh_run_command(host, base_cmd)

        abs_root = os.path.join(home, self.root)
        if helper_scripts:
            real_host = '' if self.testing else '{host}:'.format(host=host)
            cmd = "scp {helper_scripts} {host}{root}".format(
                host=real_host, root=abs_root, helper_scripts=helper_scripts
            )
            self._run_command(cmd)

        self._update_sources(host, home)

        cmd = "cd {abs_root}; ./{project_name}/.{root}/bootstrap.sh".format(
            abs_root=abs_root, root=self.root, project_name=self.project_name
        )
        try:
            self._ssh_run_command(host, cmd)
        except subprocess.CalledProcessError:
            msg = dedent("""
            ******************************************************************
            Bootstrapping of remote host {host} failed.
            All files have been copied to the host.

            Please take a look at
               {abs_root}/{project_name}/.{root}/bootstrap.sh
            and try to fix it.

            You should run it from within the {root} directory as:

               ./{project_name}/.{root}/bootstrap.sh

            Once the bootstrap.sh script runs successfully, the worker can be
            used without any further steps.

            The default bootstrap script is in
                {scripts_dir}
            and can be edited by you. These will be used for any new hosts
            you add.
            ******************************************************************
            """.format(abs_root=abs_root, root=self.root, host=host,
                       scripts_dir=self.scripts_dir,
                       project_name=self.project_name)
            )
            print(msg)
            raise BootstrapError(msg)
        else:
            print("Bootstrapping {host} succeeded!".format(host=host))

    def _get_bootstrap_code(self):
        return self.BOOTSTRAP.format(project_name=self.project_name)

    def _get_python(self, host, home):
        return os.path.join(
            home, self.root,
            'envs/{project_name}/bin/python'.format(
                project_name=self.project_name
            )
        )

    def _get_update_code(self):
        return self.UPDATE.format(project_name=self.project_name)

    def _get_helper_scripts(self):
        """Return a space separated string of script files that you need copied over to
        the remote host.

        When overriding this, you can return None or '' if you do not need any.

        """
        script = os.path.join(self.scripts_dir, 'virtualenv.pyz')
        if not os.path.exists(script):
            print("Downloading latest virtualenv.pyz")
            url = 'https://bootstrap.pypa.io/virtualenv.pyz'
            opener = urlopen(url)
            with open(script, 'wb') as f:
                f.write(opener.read())
        return script

    def _read_config(self):
        if os.path.exists(self.config_fname):
            with open(self.config_fname) as f:
                data = json.load(f)
            self.root = data['root']
            self.project_name = data['project_name']
            self.sources = data['sources']
            self.workers = data['workers']
        else:
            if self.sources is None or len(self.sources) == 0:
                project_dir = os.path.abspath(os.getcwd())
                self.project_name = os.path.basename(project_dir)
                self.sources = [project_dir]
            self.workers = [dict(host='localhost', home='', nfs=False)]
            self._write_config()
        self.scripts_dir = os.path.abspath('.' + self.root)

    def _rebuild(self, host, home):
        abs_root = os.path.join(home, self.root)
        base_cmd = "cd {abs_root}; ./{project_name}/.{root}/update.sh".format(
            abs_root=abs_root, root=self.root, project_name=self.project_name
        )
        self._ssh_run_command(host, base_cmd)

    def _run_command(self, cmd, **kw):
        print(cmd)
        subprocess.check_call(shlex.split(cmd), **kw)

    def _ssh_run_command(self, host, base_cmd):
        if self.testing:
            command = base_cmd
            print(command)
            subprocess.check_call(command, shell=True)
        else:
            command = "ssh {host} '{cmd}'".format(host=host, cmd=base_cmd)
            self._run_command(command)

    def _sync_dir(self, host, src, dest):
        options = ""
        kwargs = dict()
        if os.path.isdir(os.path.join(src, '.git')):
            exclude = 'git -C {src} ls-files --exclude-standard -oi '\
                      '--directory '.format(src=src)
            options = '--exclude-from=-'
            proc = subprocess.Popen(
                shlex.split(exclude),
                stdout=subprocess.PIPE
            )
            kwargs['stdin'] = proc.stdout
        if self.exclude_paths:
            for path in self.exclude_paths:
                options += ' --exclude="%s"' % path

        real_host = '' if self.testing else '{host}:'.format(host=host)
        command = "rsync -a {options} {src} {host}{dest} ".format(
            options=options, src=src, host=real_host, dest=dest
        )
        self._run_command(command, **kwargs)

    def _update_sources(self, host, home):
        for local_dir in self.sources:
            remote_dir = os.path.join(home, self.root + '/')
            self._sync_dir(host, local_dir, remote_dir)

        scripts_dir = self.scripts_dir
        bootstrap_code = self._get_bootstrap_code()
        update_code = self._get_update_code()
        scripts = {'bootstrap.sh': bootstrap_code,
                   'update.sh': update_code}
        for script, code in scripts.items():
            fname = os.path.join(scripts_dir, script)
            if not os.path.exists(fname):
                # Create the scripts if they don't exist.
                with open(fname, 'w') as f:
                    f.write(code)

        script_files = [os.path.join(scripts_dir, x) for x in scripts]
        for fname in script_files:
            mode = os.stat(fname).st_mode
            os.chmod(fname, mode | stat.S_IXUSR | stat.S_IXGRP)

        path = os.path.join(home, self.root, self.project_name,
                            '.' + self.root)
        real_host = '' if self.testing else '{host}:'.format(host=host)
        cmd = "scp {script_files} {host}{path}".format(
            host=real_host, path=path, script_files=' '.join(script_files)
        )
        self._run_command(cmd)

    def _delete_outputs(self, host, home, sim_dir):
        path = os.path.join(home, self.root, self.project_name,
                            sim_dir)
        real_host = '' if self.testing else '{host}'.format(host=host)
        cmd = "rm -rf {path}".format(
            path=path
        )
        self._ssh_run_command(real_host, cmd)

    def _write_config(self):
        print("Writing %s" % self.config_fname)
        data = dict(
            project_name=self.project_name,
            root=self.root,
            sources=self.sources,
            workers=self.workers
        )
        with open(self.config_fname, 'w') as f:
            json.dump(data, f, indent=2)

    # ### Public Protocol ########################################

    def add_worker(self, host, home, nfs):
        if host == 'localhost':
            self.workers.append(dict(host=host, home=home, nfs=nfs))
        else:
            curdir = os.path.basename(os.getcwd())
            if nfs:
                python = sys.executable
                chdir = curdir
            else:
                python = self._get_python(host, home)
                chdir = os.path.join(home, self.root, curdir)
            self.workers.append(
                dict(host=host, home=home, nfs=nfs, python=python, chdir=chdir)
            )

        self._write_config()
        if host != 'localhost' and not nfs:
            self._bootstrap(host, home)

    def update(self, rebuild=True):
        for worker in self.workers:
            host = worker.get('host')
            home = worker.get('home')
            nfs = worker.get('nfs', False)
            if host != 'localhost' and not nfs:
                self._update_sources(host, home)
                if rebuild:
                    self._rebuild(host, home)

    def delete(self, sim_dir, remotes):
        hosts = [h.get('host') for h in self.workers]
        if remotes[-1] == 'all':
            remotes = hosts.copy()
        else:
            for remote in remotes:
                if remote not in hosts:
                    print('%s remote is not a worker' % (remote))
                    sys.exit(1)

        for worker in self.workers:
            host = worker.get('host')
            home = worker.get('home')
            nfs = worker.get('nfs', False)
            if host in remotes:
                if host != 'localhost' and not nfs:
                    self._delete_outputs(host, home, sim_dir)


    def create_scheduler(self):
        """Return a `automan.jobs.Scheduler` from the configuration.
        """
        from .jobs import Scheduler

        scheduler = Scheduler(root='.')
        for worker in self.workers:
            host = worker.get('host')
            nfs = worker.get('nfs', False)
            if host == 'localhost':
                scheduler.add_worker(dict(host='localhost'))
            else:
                python = worker.get('python')
                chdir = worker.get('chdir')
                config = dict(host=host, python=python, chdir=chdir, nfs=nfs)
                if self.testing:
                    config['testing'] = True
                scheduler.add_worker(config)
        return scheduler

    def cli(self, argv=None):
        """This is just a demonstration of how this class could be used.
        """
        import argparse
        parser = argparse.ArgumentParser(description='Setup remote workers.')

        parser.add_argument(
            '-a', '--add-node', action="store", dest="host", type=str,
            default='', help="Add a new remote worker."
        )
        parser.add_argument(
            '--home', action="store", dest="home", type=str,
            default='',
            help='Home directory of the remote worker (to be used with -a)'
        )
        parser.add_argument(
            '--nfs', action="store_true", dest="nfs",
            default=False,
            help=('Does the remote remote worker share the filesystem '
                  '(to be used with -a)')
        )
        parser.add_argument(
            '--no-rebuild', action="store_true", dest="no_rebuild",
            default=False, help="Do not rebuild the sources on sync."
        )

        args = parser.parse_args(argv)

        if len(args.host) == 0:
            self.update(not args.no_rebuild)
        else:
            self.add_worker(args.host, args.home, args.nfs)



from automan.api import Problem, Automator


class Squares(Problem):
    def get_name(self):
        return 'squares'

    def get_commands(self):
        return [
            ('1', 'python square.py 1', None),
            ('2', 'python square.py 2', None),
            ('3', 'python square.py 3', None),
            ('4', 'python square.py 4', None),
        ]

    def run(self):
        self.make_output_dir()
        data = []
        for i in ('1', '2', '3', '4'):
            stdout = self.input_path(i, 'stdout.txt')
            with open(stdout) as f:
                data.append(f.read().split())

        output = self.output_path('output.txt')
        with open(output, 'w') as o:
            o.write(str(data))


if __name__ == '__main__':
    automator = Automator(
        simulation_dir='outputs',
        output_dir='manuscript/figures',
        all_problems=[Squares]
    )

    automator.run()
